# WATCHDOG APP FOR USER-SPECIFIED BLUE IRIS LOGFILE ENTRIES
# IF AN EVENT IS DETECTED...
# PUSH A NOTIFICATION VIA THE PUSHOVER API OR TELEGRAM API
#
# This script monitors the active Blue Iris logfile for new entries that match a user-specified filter (or list of filters).
#
# Version: 3
# Created by jaydeel at ipcamtalk.com
#
# Updates:
# V3 - 2025-02-22 - BREAKING CHANGE: must replace old user-settings file
#                 - FIXED: highlighted found terms in the notification message now preserves the logfile entry's original capitalization
#                 - ADDED: argument -PushService (and user-setting $push_service) may now be 'all' as well as 'Pushover' or 'Telegram'
#                 - CHANGED: user setting $use_bi_creds_flg to $post_w_creds_flg
#                 - IMPROVED: updated functions SendPushover() & SendTelegram()
# V2 - 2024-10-26 - FIXED: formating of summary string when using comma 'AND' logical operators
# V1 - 2024-10-25 - initial release
#
# USAGE:      {this script name}.ps1 [arg1]
# ARGUMENTS:  -FilterMsgs    = list of filter terms in the 'message' column; comma-delimited"
#             -FilterLevels  = list of filter terms in the 'level' column; comma-delimited"
#             -FilterObjects = list of filter terms in the 'object' column; comma-delimited"
#             -BatchFile     = path to *.txt file (alternate to -FilterMsgs, -FilterLevels and -FilterObjects) [3]"F
#             -Reset         = (switch) clear temp files used to recursively execute script via -BatchFile"
#             -Renew         = (switch) clear temp files used to track number of new rows added to the logfile"
#             -ShowMatches   = (switch) show filtered results in console
#             -PushService   = (string) 'Pushover', 'Telegram', or 'all'; overrides user-settings '$push_service'
#             -Title         = (string) on-demand title for notification
#             -Priority      = (int) on-demand priority for notification (Pushover only)
#             -Sound         = (string) on-demand sound for notification (Pushover only)
#             -Devices       = (string) on-demand devices for notification (Pushover only)F
#             -NoPush        = (switch) disables sending push service notification (enabled by default)
#             -NoBiLog,      = (switch) disable/override logging to the Blue Iris logfile
#             -Quiet         = suppresses messaging in the background, and verbose informational output in the console
#             -Help          = show usage guide
#             -Examples      = show examples usage guide
#             -DevMode       = (dev) show developer mode info
#             -TestFile      = (dev) use a mock *.txt file instead of the active logfile
# RETURNS:    (array) found new entries from Blue Iris active logfile
# NOTES:      Script accesses the Windows registry to automatically retrieve: Blue Iris logfile path, and webserver IP & port
#
# BEFORE USING THIS SCRIPT:
#   1. EDIT ALL VARIABLES IN THE 'USER SETTINGS' FILE
#
# HOW TO AUTOMATE THIS SCRIPT VIA TASK SCHEDULER
#   Open Task Scheduler and create an appropriately named task with one trigger and one action:
#   Task Scheduler
#   'Triggers' tab:
#       'At startup', 'Delay task for' (e.g., 30 minutes)  <-- use if delaying Blue Iris startup for CPAI, etc.
#                     'Repeat task every' (e.g., 30 minutes)
#   'Actions' tab:
#       Program/script: powershell.exe
#       Add arguments: {your_powershell_scripts_path)\{script_file_name}.ps1 [args]
#
# DEVELOPER'S NOTE
#   (2/16/2025) The Task Scheduler Task that is executing this script is currently being monitored by 
#   another Task Scheduler Task that is running script 'critical_scheduled_tasks_wdog.ps1' (every 4 hours).
#   This script: Library > My Blue Iris > 'BI Memleak Wdog V2' 
#   Companion script: Library > My Tasks > 'Critical Tasks Watchdog' 

param (
	[String[]]$FilterMsgs,    # analyze specified items in 'message' column; comma-delimited
	[String[]]$FilterLevels,  # analyze specified items in 'level' column (0,1,2); comma-delimited
	[String[]]$FilterObjects, # analyze specified items in 'object' column; comma-delimited
	[String]$BatchFile,       # path to file (alternate to -FilterMsgs, -FilterLevels and -FilterObjects)
	[Switch]$Reset,           # clear temp files used to recursively execute script via -BatchFile"
	[Switch]$Renew,           # clear temp files used to track number of new rows added to the logfile
	[Switch]$ShowMatches,     # show filtered results in console
	[String]$PushService,     # 'Pushover', 'Telegram', or 'all'; overrides user-settings-file '$push_service'
	[String]$Title,           # on-demand title for notification
	[Int]$Priority,           # on-demand priority for notification (Pushover only)
	[String[]]$Sound,           # on-demand sound for notification (Pushover only)
	[String[]]$Devices,         # on-demand devices for notification (Pushover only)
	[Switch]$NoPush,          # do not send push service notification (useful while testing)
	[Switch]$NoBiLog,         # disable/override logging to the Blue Iris logfile (useful while testing)
	[Switch]$Quiet,           # suppress verbose informational output to the console
	[Switch]$Help,            # show usage guide
	[Switch]$Examples,        # display script usage examples
	[Switch]$DevMode,         # (dev) show developer mode info
	[String]$TestFile         # (dev) use this file instead of the active logfile
)

# Set-StrictMode -Version 3  # https://tinyurl.com/mrx3w9v4
# Start-Transcript  # default destination is C:\Users\{user}l\Documents
# $ErrorActionPreference = "Stop"

# For debugging
# Write-Host "Entered script..."
# Write-Host ("FilterMsgs type = " + $FilterMsgs.GetType().Name)
# Write-Host ("FilterMsgs cnt = " + $FilterMsgs.Count)
# Write-Host "FilterMsgs = $FilterMsgs"
# Write-Host ("FilterMsgs = " + ($FilterMsgs | ConvertTo-Json))
# Write-Host ("FilterLevels cnt = " + $FilterLevels.Count)
# Write-Host "FilterLevels = $FilterLevels"
# Write-Host ("FilterLevels = " + ($FilterLevels | ConvertTo-Json))
# Write-Host ("FilterObjects cnt = " + $FilterObjects.Count)
# Write-Host ("FilterObjects = " + ($FilterObjects | ConvertTo-Json))
# Write-Host "FilterObjects = $FilterObjects"
# Write-Host "BatchFile = $BatchFile"
# Write-Host "ShowMatches = $ShowMatches"
# Write-Host "PushService = $PushService"
# Write-Host "Title = '$Title'"
# Write-Host "Priority = $Priority"
# Write-Host "Sound = $Sound"
# Write-Host "Devices = $Devices"
# Write-Host "NoPush = $NoPush"
# Write-Host "DevMode = $DevMode"
# Write-Host "TestFile = $TestFile"

$ScriptVersion = "V3"

# FUNCTIONAL DESCRIPTION
# Check if using argument -BatchFile
#   * Process the BatchFile & create batch management files in $app_path\_script_files
# Check for the existence of batch management files and if running the 2nd or higher actionable line in the BatchFile
#   * Proceed as 'recursive mode'
#      * Reuse most recent entries from the Blue Iris logfile
# Otherwise
#   * Proceed in normal mode
#     * Clear any existing batch management files in $app_path\_script_files
#     * Import the most recent entries from the Blue Iris logfile
#     * Save the entries and count to files 'active_logfile_rows_*.txt' in $app_path
# Process the 'filter set' (arguments: -FilterLvels, -Filter-Objects and/or -FilterMsgs)
# If matches found
#   * Send notification
#   * Post filter set results summary to the Blue Iris logfile (optional)
#   * Append filter set results summary to script's CSV file
# Check if in recursive mode and if this is not the final actionable line in -BatchFile
#   * Update the batch management files in $app_path\_script_files
#   * Run the script again with the next actionable line the -BatchFile
# Terminate script
#   * Note: batch management files are not deleted to enable inspection for debugging, if necessary

# IMPORT USER SETTINGS...
# Note: The 'user settings' script is expected to be in same folder as this script
$ScriptName = $MyInvocation.MyCommand.Name
$user_settings_filename = $ScriptName.Replace(".ps1", "_user_settings.ps1")
try {
	. $user_settings_filename
	
	# For debugging
	#Write-Host "Loaded user settings from '$user_settings_filename'"
}
catch {
	return "ERR/failed to load user settings file '$user_settings_filename'`nTip: try executing it directly to confirm it exists and reveal syntax errors"
}

# ------------------------------------
# VALIDATE USER SETTINGS
if (1 -eq 1) {  # Use to collapse this code section while editing

	# REQUIRED USER SETTINGS
	
	$variable_test = Get-Variable -Name app_path -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$app_path"
		return $err_msg
	} else {
		if ($app_path.StartsWith('YOUR')) {
			$err_msg = "ERR/user-setting file setting `$app_path is not set"
			return $err_msg
		}
	}
	$app_path = $app_path.Replace("/","\").TrimEnd('\')

	$variable_test = Get-Variable -Name my_default_filters -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$my_default_filters = @()
	}
	$using_my_default_filters_flg = ($my_default_filters.count -gt 0)

	$variable_test = Get-Variable -Name bi_user_pw -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/User-settings file is missing setting `$bi_user_pw"
		return $err_msg
	} else {
		if ($bi_user_pw.StartsWith('YOUR')) {
			$err_msg = "ERR/user-setting file setting `$bi_user_pw is not set"
			return $err_msg
		}
	}

	$variable_test = Get-Variable -Name push_service -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$push_service = 'pushover'
	} else {
		if (!(@('pushover', 'telegram', 'all') -contains $push_service.ToLower())) {  # -contains operator is case-insensitive by default
			$err_msg = "ERR/user-setting file setting `$push_service must be 'Pushover', 'Telegram', or 'all'"
			return $err_msg
		}
	}
	$push_service = $push_service.ToLower()

	$variable_test = Get-Variable -Name push_row_matches_disp_max -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$push_row_matches_disp_max = 3
	}
	
	if ($push_service -eq 'pushover' -or $PushService -match 'pushover' -or $push_service -match 'all') {

		$variable_test = Get-Variable -Name po_user_token -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user-settings file is missing setting `$po_user_token"
			return $err_msg
		} else {
			if ($po_user_token.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting `$po_user_token' is not set"
				return $err_msg
			}
		}

		$variable_test = Get-Variable -Name po_app_token -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user-settings file is missing setting `$po_app_token"
			return $err_msg
		} else {
			if ($po_app_token.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting `$po_app_token is not set"
				return $err_msg
			}
		}
	}
	
	if ($push_service -match 'telegram' -or $PushService -match 'telegram' -or $push_service -match 'all') {

		$variable_test = Get-Variable -Name tg_api_token -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user-settings file is missing setting `$tg_api_token"
			return $err_msg
		} else {
			if ($tg_api_token.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting `$tg_api_token' is not set"
				return $err_msg
			}
		}

		$variable_test = Get-Variable -Name tg_chat_id -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user-settings file is missing setting `$tg_chat_id"
			return $err_msg
		} else {
			if ($tg_chat_id.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting `$tg_chat_id is not set"
				return $err_msg
			}
		}
	}

	$variable_test = Get-Variable -Name post_w_creds_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_w_creds_flg = $false
	}
	# Set global var used by PostCustomBiLogEntry()
	$post_w_creds_flg = $post_w_creds_flg  

	if ($post_w_creds_flg) {
		
		$variable_test = Get-Variable -Name bi_user_name -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/User-settings file is missing setting `$bi_user_name"
			return $err_msg
		} else {
			if ($bi_user_name.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting `$bi_user_name is not set"
				return $err_msg
			}
		}
		# Write-Host "bi_user_name = $bi_user_name"

		$variable_test = Get-Variable -Name bi_user_pw -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/User-settings file is missing setting `$bi_user_pw"
			return $err_msg
		} else {
			if ($bi_user_pw.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting `$bi_user_pw is not set"
				return $err_msg
			}
		}
	}

	# OPTIONAL USER SETTINGS
	
	<# $variable_test = Get-Variable -Name experienced_user_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$experienced_user_flg = $false
	} #>
	
	$variable_test = Get-Variable -Name bg_msgs_on -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$bg_msgs_on = $true
	}
	
	$variable_test = Get-Variable -Name post_to_BI_logfile_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_to_BI_logfile_flg = $false
	}
	
	$variable_test = Get-Variable -Name custom_BI_logfile_level -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$custom_BI_logfile_level = 0  # 'informational'
	}
	
	$variable_test = Get-Variable -Name log_to_csv_logfile_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$log_to_csv_logfile_flg = $true
	}
	
	$variable_test = Get-Variable -Name log_only_events_to_csv_logfile_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$log_only_events_to_csv_logfile_flg = $false
	}

	$variable_test = Get-Variable -Name post_to_BI_logfile_w_matches_only_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_to_BI_logfile_w_matches_only_flg = $false
	}
	
	$variable_test = Get-Variable -Name post_to_BI_logfile_in_background_only_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_to_BI_logfile_in_background_only_flg = $false
	}

}

# GATHER/VALIDATE SCRIPT & ENVIRONMENT INFO
if (1 -eq 1) {  # Use to collapse this code section while editing

	#$ScriptName = $MyInvocation.MyCommand.Name
	#$ScriptName = Split-Path -leaf $PSCommandpath
	$ScriptPath = $PSCommandPath.Replace("\$ScriptName", '')
	$ScriptBaseName = $ScriptName.Replace('.ps1', '')
	$ScriptPath = $PSCommandPath.Replace("\$ScriptName", '')

	# For debugging
	#Write-Host "ScriptName = $ScriptName"
	# Write-Host "ScriptPath = $ScriptPath"
	# Write-Host "ScriptBaseName = $ScriptBaseName"
	# Write-Host "ScriptPath = $ScriptPath"

	# INITIALIZE VARS
	$PsVersion = $PSVersionTable.PSVersion.Major

	# Task Scheduler is 'svchost', Powershell-ISE is 'explorer', Windows Terminal is 'WindowsTerminal'
	$ParentProcess = (Get-Process -Id (Get-CimInstance Win32_Process -Filter "ProcessID = $pid").ParentProcessId).Name
	# ... create lookup table (used by script's CSV log)
	$HostLabels = @{}
	$HostLabels['svchost'] = 'TaskScheduler'
	$HostLabels['explorer'] = 'Powershell or Powershell-ISE'
	$HostLabels['WindowsTerminal'] = 'WindowsTerminal'

	# For testing
	#Write-Host "ParentProcess = $ParentProcess"

	# ... Get friendly host name
	if ($HostLabels[$ParentProcess]) {
		$script_host_str = $HostLabels[$ParentProcess]
	} else {
		$script_host_str = $ParentProcess
	}
	# Write-Host "script_host_str = $script_host_str"

	# NOTE: Scripts called by another script have a 'PSCommandPath'.
	#       Method 'PSCommandPath' returns null if script is executed via the terminal
	#       https://stackoverflow.com/a/23370618/13499984
	$CallingScript = $MyInvocation.PSCommandPath

	# DETERMINE IF RUNNING IN BACKGROUND (IN CLI MODE, OR FROM TASK SCHEDULER, BLUE IRIS, etc.)
	$FromConsole = $False  # Default
	if ($null -eq $CallingScript) {
		$FromConsole = $True
	} elseif ($CallingScript -match $ScriptName) {  # Detect recursive calls
		$FromConsole = $True
	}
	
	# Save the console's original foreground color
	if ($FromConsole) {
		$OriginalForegoumdColor = $Host.UI.RawUI.ForegroundColor
	}

	# Get the current username and domain (hostname)
	# NOTE: sometimes needed for msg.exe (e.g. from PowerShell ISE)
	# ... tip: from the console, you can type 'hostname' or 'whoami'
	$my_servername = (whoami).Split('\')[0]
	$server_username = (whoami).Split('\')[1]

	# For debugging
	# Write-Host "FromConsole = $FromConsole"
	# msg.exe $server_username /TIME:10 "entered script '$ScriptName'"
	# msg.exe $server_username /TIME:10 "PsVersion ($ScriptName) = $PsVersion"
	# msg.exe $server_username /TIME:10 "ParentProcess ($ScriptName) = $ParentProcess"
	# msg.exe $server_username /TIME:10 "FromConsole ($ScriptName) = $FromConsole"
	# msg.exe $server_username /TIME:5 "From script... $ScriptName`n-----`nPowerShell version is... $PsVersion`nParent process is... '$ParentProcess'`nCalling script name... $CallingScript"
	# Write-Host "CallingScript = $CallingScript"

	# SHOW OPTIONAL MESSAGE WHEN RUNNING FROM BACKGROUND
	if ($bg_msgs_on -and !$FromConsole) {
		$ExecPolicy = Get-ExecutionPolicy
		$exec_policy_str = "PSExecutionPolicy = $ExecPolicy"
		if ($env:PSExecutionPolicyPreference) {
			$ExecPolicyPref = $env:PSExecutionPolicyPreference
			$exec_policy_str = "Session ExecutionPolicy = $ExecPolicyPref"
		}
		$msg_sec = 10
		$bg_msg = "Script '$ScriptName' is running in the background" + 
		"`n$exec_policy_str" + 
		"`nPsVersion = $PsVersion" + 
		"`nParentProcess = $ParentProcess" + 
		"`n`n(this msg auto-closes in $msg_sec seconds)"

		msg.exe $server_username /TIME:$msg_sec $bg_msg
	}

	# For testing - Simulate CLI mode
	#$FromConsole = $false

	# FEEDBACK UNEXPECTED COMMANDLINE ARGUMENTS
	#Write-Host ("params = " + ($PSBoundParameters | ConvertTo-Json))
	$cmdline_args_passed = [String[]]$PSBoundParameters.Keys
	#Write-Host "Extra Args w/ values = $Args"
	$cmdline_args_bad = ($Args | where{$_ -like "-*"}) | ForEach-Object{$_ -Replace '-', ''}
	if ($cmdline_args_bad) {
		$cmdline_args_bad = [String[]]$cmdline_args_bad
	} else {
		$cmdline_args_bad = @()
	}

	# For debugging
	# Write-Host "Cmdline Args Passed = $cmdline_args_passed"
	# Write-Host "Cmdline Args Bad = $cmdline_args_bad"
	# Write-Host ("Cmdline Args Passed = " + ($cmdline_args_passed | ConvertTo-Json))
	# Write-Host ("Cmdline Args Bad = " + ($cmdline_args_bad | ConvertTo-Json))

	if ($cmdline_args_bad.Count -gt 0) {
		if ($FromConsole) {
			Write-Host ("HEADS-UP: commandline contains unsupported argument(s): [ " + ($cmdline_args_bad -Join ', ') + " ]") -ForegroundColor Green
			return
		}
	}
}

# ==============================================================

# FUNCTIONS
if (1 -eq 1) {  # Use to collapse this code section while editing

	function ConvertTo-Hashtable() {
	# PURPOSE:  To take an object returned from the 'ConvertFrom-Json' CmdLet and convert it to a Powershell hashtable.
	# NOTE:     PowerShell has a native command called ConvertFrom-Json, but (before PowerShell 7) 
	#           it did not directly convert data to a hashtable
	# Source:   https://4sysops.com/archives/convert-json-to-a-powershell-hash-table/

		[CmdletBinding()]
		[OutputType('hashtable')]
		param (
			[Parameter(ValueFromPipeline)]
			$InputObject
		)

		process {
			# Return null if the input is null. This can happen when calling the function
			# recursively and a property is null
			if ($null -eq $InputObject) {
				return $null
			}

			# Check if the input is an array or collection. If so, we also need to convert
			# those types into hash tables as well. This function will convert all child
			# objects into hash tables (if applicable)
			if ($InputObject -is [System.Collections.IEnumerable] -and $InputObject -isnot [String]) {
				$collection = @(
					foreach ($object in $InputObject) {
						ConvertTo-Hashtable -InputObject $object  # RECURSIVE CALLS
					}
				)
				# Return the array but don't enumerate it because the object may be pretty complex
				Write-Output -NoEnumerate $collection
			} elseif ($InputObject -is [psobject]) { ## If the object has properties that need enumeration
				# Convert it to its own hash table and return it
				$hash = @{}
				foreach ($property in $InputObject.PSObject.Properties) {
					$hash[$property.Name] = ConvertTo-Hashtable -InputObject $property.Value
				}
				$hash
			} else {
				# If the object isn't an array, collection, or other object, it's already a hash table
				# So just return it.
				$InputObject
			}
		}
	}

	function RegGetValue() {
	# PURPOSE:  Get current value of a registry key
	# ARGS:     see param lines below
	# RETURNS:  registry value, user-provided default value, or error message
	# 
	# V2 (2024/07/01)
	# Created by jaydeel @ ipcamtalk.com

		param(
			[String]$KeyPath,  # like, 'HKLM:SOFTWARE\Perspective Software\Blue Iris\clips'
			[String]$KeyName,  # like, 'compact'
			$DefaultValue      # value to return if $KeyName does not exist 
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
		
		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host "$fn_name() -KeyPath = '$KeyPath'"
		# Write-Host "$fn_name() -KeyName = '$KeyName'"
		# Write-Host "$fn_name() -DefaultValue = '$DefaultValue'"


		try {
			$reg_obj = Get-ItemProperty -Path $KeyPath -Name $KeyName -ErrorAction Stop
		
			# For debugging
			#Write-Host ("reg_obj = " + ($reg_obj | ConvertTo-Json -Depth 100))
			
			if ($reg_obj) {
				if (!($null -eq $reg_obj.$KeyName)) {
					return $reg_obj.$KeyName
				}
			}
		}
		catch {
			# continue
		}	

		if ($DefaultValue) {
			$rtn_value = $DefaultValue
		} else {
			$err_msg = "ERR/unable to get registry key '$KeyPath\$KeyName'"
			$rtn_value = $err_msg
		}
		
		return $rtn_value
	}

	function Convert-HexToVersion() {
	# PURPOSE: To convert a hexadecimal number like 5090704 or 5091014 to a string like 5.9.7.4 or 5.9.10.14
	# Source: ChatGPT (modified by jaydeel)

		param (
			[Parameter(Mandatory=$true)]
			[String]$Hex
		)
	 
		# For debugging
		# Write-Host "Entered $fn_name()..."
		# Write-Host "$fn_name() Hex = '$Hex'"
		
		# Ensure the hex number is a string (if given as a number)
		$HexString = $Hex.ToString()

		# Ensure the hex string has an even number of characters by prepending a '0' if needed
		if ($HexString.Length % 2 -ne 0) {
			$HexString = "0" + $HexString
		} 

		# Split the hex string into pairs of characters
		# ... inserts a period after every two characters in the string
		$pairs = for ($i = 0; $i -lt $hex.Length; $i += 2) {
			$HexString.Substring($i, 2)
		}
		
		# For debugging
		# Write-Host ("$fn_name() pairs = " + ($pairs | ConvertTo-Json))

		# Convert each pair to its decimal equivalent and join them with '.'
		return ($pairs | ForEach-Object { [convert]::ToInt32($_, 16) }) -Join '.'
	}

	function IsBiVersion() {
	# PURPOSE:   To confirm if the CURRENT Blue Iris version exceeds a specified UPDATE version
	# ARGUMENTS: -CompareToVersion = [REQUIRED] (string) compare version, '5.7.2.3'
	#            Note: supports version strings 'X.000.000.000' to 'X.999.999.999'            
	# RETURNS:   boolean; $true current BI version exceeds specified version
	# REQUIRES:  function RegGetValue(), Convert-HexToVersion()
		
		param (
			[String]$CompareToVersion
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# Write-Host "Entered $fn_name()..."
		# Write-Host "$fn_name() UpdateVersion = $CompareToVersion"

		# Start timer
		# $timer = [Diagnostics.Stopwatch]::StartNew()
		
		$rtn_flg = $false  # Assume failure

		# GET CURRENT VERSION
		
		# Get current setting's value
		$key_path = 'HKLM:\SOFTWARE\Perspective Software\Blue Iris'
		$key_name = 'version'
		$key_val_cur = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name

		# Abort if GET failed
		if ($key_val_cur -match 'ERR/') {
			if ($script:FromConsole) {
				Write-Host "$fn_name() ERR/REG request failed"
				Write-Host "$fn_name() RegGetValue() key_val_cur = $key_val_cur"
			}
			return $false
		}

		# Get the current version string
		$bi_version_hex = $key_val_cur.ToString("X")  # like hex 5090704
		$bi_version = Convert-HexToVersion -Hex $bi_version_hex  # like '5.9.7.4'
		
		# Make comparator integer arrays
		[int[]]$my_ver_ary = $CompareToVersion.Split('.')
		[int[]]$bi_ver_ary = $bi_version.Split('.')

		# Make 'prime.subprime' version comparators, like 5.8 vs 5.7
		$my_ver_prime_dot = $my_ver_ary[0] + $my_ver_ary[1]/1000
		$bi_ver_prime_dot = $bi_ver_ary[0] + $bi_ver_ary[1]/1000

		# For debugging
		# Write-Host "bi_version_hex (registry) = $bi_version_hex"
		# Write-Host "bi_version = $bi_version"
		# Write-Host ("my_ver_ary = " + ($my_ver_ary | ConvertTo-Json))
		# Write-Host ("bi_version_ary = " + ($bi_ver_ary | ConvertTo-Json))
		# Write-Host "my_ver_prime_dot = $my_ver_prime_dot"
		# Write-Host "bi_ver_prime_dot = $bi_ver_prime_dot"
		
		if ($my_ver_prime_dot -eq $bi_ver_prime_dot) {

			# Make sub-subprime version comparators, 5.001 for 5.8.5.1, 999.020 for 5.8.999.20
					
			$mult1 = [Math]::Pow(10, 0)
			$mult2 = [Math]::Pow(10, -3)

			if ($my_ver_ary[2] -eq 0 -and $my_ver_ary[3] -eq 0) {
				$my_ver_sub = 0.000
			} else {
				$my_ver_sub = $my_ver_ary[2]*$mult1 + $my_ver_ary[3]*$mult2
			}

			if ($bi_ver_ary[2] -eq 0 -and $bi_ver_ary[3] -eq 0) {
				$bi_ver_sub = 0.000
			} else {
				$bi_ver_sub = $bi_ver_ary[2]*$mult1 + $bi_ver_ary[3]*$mult2
			}

			# For debugging
			# Write-Host ("my_ver_sub = " + ("{0:n3}" -f $my_ver_sub))
			# Write-Host ("bi_ver_sub = " + ("{0:n3}" -f $bi_ver_sub))

			$rtn_flg = ($bi_ver_sub -ge $my_ver_sub) 


		} else {
			$rtn_flg = ($bi_ver_prime_dot -ge $my_ver_prime_dot)
		}

		# Stop timer
		# $timer.stop()
		# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
		# $elapsed_msec_str = $elapsed_msec.ToString("0.00")
		# Write-Host ("$fn_name() timer = $elapsed_msec_str msec")

		return $rtn_flg
	}

	function GetBiWebserverIpPort() {
	# PURPOSE:   Get the Blue Iris web server's LAN IP and Port
	# ARGUMENTS: None           
	# RETURNS:   string, like "192.168.1.3:81"
	# REQUIRES:  function RegGetValue()
	# 
	# V1 (9/30/2024)
	# by jaydeel at ipcamtalk.com
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# Write-Host "Entered $fn_name()..."

		# Start timer
		# $timer = [Diagnostics.Stopwatch]::StartNew()
		
		$rtn_str = ""  # Assume failure
		
		# 1) GET LAN IP VALUE
		$key_path = 'HKLM:\SOFTWARE\Perspective Software\Blue Iris\server'
		$key_name = 'lanip'
		$key_val_cur = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name

		# Abort if GET failed
		if ($key_val_cur -match 'ERR/') {
			if ($script:FromConsole) {
				Write-Host "$fn_name() ERR/REG request failed"
				Write-Host "$fn_name() RegGetValue() key_val_cur = $key_val_cur"
			}
			return $rtn_str
		}
		
		$server_lan_ip = $key_val_cur
		
		# 2) GET WEB SERVER PORT VALUE
		$key_path = 'HKLM:\SOFTWARE\Perspective Software\Blue Iris\server'
		$key_name = 'port'
		$key_val_cur = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name

		# Abort if GET failed
		if ($key_val_cur -match 'ERR/') {
			if ($script:FromConsole) {
				Write-Host "$fn_name() ERR/REG request failed"
				Write-Host "$fn_name() RegGetValue() key_val_cur = $key_val_cur"
			}
			return $rtn_str
		}

		$server_port = $key_val_cur

		# Stop timer
		# $timer.stop()
		# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
		# $elapsed_msec_str = $elapsed_msec.ToString("0.00")
		# Write-Host ("$fn_name() timer = $elapsed_msec_str msec")
		
		# Build the IP & Port address
		$rtn_str = $server_lan_ip + ":" + $server_port

		return $rtn_str
	}

	function GetBiLogPath() {
	# PURPOSE:   Get the Blue Iris log path
	# ARGUMENTS: None           
	# RETURNS:   string, like "c:\BlueIris\log"
	# REQUIRES:  function RegGetValue()
	# 
	# V1 (10/7/2024)
	# by jaydeel at ipcamtalk.com
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# Write-Host "Entered $fn_name()..."

		# Start timer
		# $timer = [Diagnostics.Stopwatch]::StartNew()
		
		$rtn_str = ""  # Assume failure

		# GET THE LOG FILE'S PATH TEMPLATE
		# ... expecting a sting like 'c:\BlueIris\log\%Y%W.txt'
		$key_path = 'HKLM:\SOFTWARE\Perspective Software\Blue Iris\options'
		$key_name = 'LogPath'
		$key_val_cur = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name

		# Abort if GET failed
		if ($key_val_cur -match 'ERR/') {
			if ($script:FromConsole) {
				Write-Host "$fn_name() ERR/REG request failed"
				Write-Host "$fn_name() RegGetValue() key_val_cur = $key_val_cur"
			}
			return $rtn_str
		}

		# Extract the path
		#$rtn_str = Resolve-Path -Path $key_val_cur  # fails if filespec does not exist
		$rtn_str = Split-Path -Path $key_val_cur -Parent

		# Stop timer
		# $timer.stop()
		# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
		# $elapsed_msec_str = $elapsed_msec.ToString("0.00")
		# Write-Host ("$fn_name() timer = $elapsed_msec_str msec")

		return $rtn_str
	}

	function GetBiWebserverUserPw() {
	# PURPOSE:   Get the password for a Blue Iris web server user name
	#            UNDER DEVELOPMENT - NOT YET WORKING
	# ARGUMENTS: -UserName           
	# RETURNS:   string
	# REQUIRES:  function RegGetValue()
	# 
	# V1 (10/7/2024)
	# by jaydeel at ipcamtalk.com
		
		param (
			[String]$UserName
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name()..."
		# Write-Host "$fn_name() UserName = $UserName"

		# Start timer
		# $timer = [Diagnostics.Stopwatch]::StartNew()
		
		$rtn_str = ""  # Assume failure

		# GET THE USER NAME'S ENCRYPTED PASSWORD
		$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\server\users\$UserName"
		$key_name = 'password'
		$key_val_cur = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name

		# Abort if GET failed
		if ($key_val_cur -match 'ERR/') {
			if ($script:FromConsole) {
				Write-Host "$fn_name() ERR/REG request failed"
				Write-Host "$fn_name() RegGetValue() key_val_cur = $key_val_cur"
			}
			return $rtn_str
		}

		# DECRYPT THE PASSWORD

		# Make sure is the reg key value is a valid base64 string
		try {
			$encryptedBytes = [Convert]::FromBase64String($key_val_cur)
			# Write-Host "Base64 decoded successfully."
		} catch {
			# Write-Host "Error decoding base64 string: $_"
			return "ERR/error decoding base64 string: $_"
		}
		
		# Convert the base64-encoded string back to a byte array
		$encryptedBytes = [Convert]::FromBase64String($key_val_cur)
		# Write-Host "encryptedBytes = $encryptedBytes"

		$entropy = $null  # No entropy used
		#$entropy = [System.Text.Encoding]::UTF8.GetBytes("your_entropy_here")  # Replace with your entropy
		$entropy = [System.Text.Encoding]::UTF8.GetBytes("BlueIris")

		try {
			$decryptedBytes = [System.Security.Cryptography.ProtectedData]::Unprotect(
				$encryptedBytes, 
				$entropy,  # Same entropy used during encryption
				[System.Security.Cryptography.DataProtectionScope]::CurrentUser
			)
			$decryptedString = [System.Text.Encoding]::UTF8.GetString($decryptedBytes)
			Write-Host "Decrypted string: $decryptedString"
			$rtn_str = $decryptedString
		} catch {
			Write-Host "Decryption error: $_"
		}

		# Stop timer
		# $timer.stop()
		# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
		# $elapsed_msec_str = $elapsed_msec.ToString("0.00")
		# Write-Host ("$fn_name() timer = $elapsed_msec_str msec")

		return $rtn_str
	}

	function FormatLogEntriesForPush() {
	# PURPOSE:   Format Blue Iris logfile entries in a concise format suitable for push messages
	# ARGUMENTS: see param line below            
	# RETURNS:   string
	# REQUIRES:  nothing

		param (
			[String[]]$Rows,
			[Int]$RowsMax,
			[String[]]$UsedLevels,
			[String[]]$UsedTerms,
			[hashtable]$PureOrQueryHt  # hashtable for single filters that have only 'OR' logical operators
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name()..."
		# Write-Host ("$fn_name() Rows type = " + $Rows.GetType().Name)
		# Write-Host ("$fn_name() Rows cnt = " + $Rows.Count)
		# Write-Host ("$fn_name() Rows = " + ($Rows | ConvertTo-Json))
		# Write-Host "$fn_name() RowsMax = $RowsMax"
		# Write-Host ("$fn_name() UsedLevels = " + ($UsedLevels | ConvertTo-Json))
		# Write-Host ("$fn_name() UsedTerms = " + ($UsedTerms | ConvertTo-Json))
		# Write-Host ("$fn_name() PureOrQueryHt type = " + $PureOrQueryHt.GetType().Name)
		# Write-Host ("$fn_name() PureOrQueryHt keys = " + $PureOrQueryHt.Keys)
		# Write-Host ("$fn_name() PureOrQueryHt = " + ($PureOrQueryHt | ConvertTo-Json))
		
		$rtn_str = ""

		if ($PureOrQueryHt.Keys.Count -eq 0) {
			
			# TYPICAL CASE
			
			$max_rows = $RowsMax
			$lines = $Rows
			# KEY POINT: Reversing the array facilitates returning the most recent entries
			[Array]::Reverse($lines)
		
		} else {

			# SPECIAL CASE - AN ALL 'OR' QUERY
			# like, -FilterMsgs "keep up'|force|fast enough|cannot|can't|unexpected|not responding|file exists"

			# Expecting PureOrQueryHt to be like,
			#	{
			#	  "fast enough": [
			#	    "2\t8/22/2024 12:01:43.615 AM\tDW1\tClip: Disk can't write fast enough (0 2594 30010 74024)",
			#	    "2\t8/22/2024 12:02:06.131 AM\tFD\tClip: Disk can't write fast enough (0 1918 30005 95802)",
			#	    "2\t10/20/2024 12:47:24.777 PM\tBY\tClip: Disk can't write fast enough (0 1920 30010 0)"
			#	  ],
			#	  "not responding": [
			#	    "2\t10/20/2024 10:47:22.607 AM\tDW1\tAI: not responding",
			#	    "2\t10/20/2024 10:47:23.022 AM\tDW2\tAI: not responding",
			#	    "2\t10/20/2024 10:47:23.544 AM\tDW2\tAI: not responding",
			#	    "2\t10/20/2024 10:47:24.044 AM\tDW2\tAI: not responding",
			#	    "2\t10/20/2024 10:47:24.777 AM\tDW2\tAI: not responding"
			#	  ]
			#	}
			
			# Make sure the output shows AT LEAST ONE line for each matched query item (i.e., each key in $PureOrQueryHt)
			# KEY POINT: each key in $PureOrQueryHt should have an array w/ AT LEAST ONE entry
			$max_rows = $PureOrQueryHt.Keys.Count
			
			# REORGANIZE THE DATA		
			if (1 -eq 1) {  # Use to collapse this code section while editing

				# Initialize hashtable collector
				$orderedHt = [System.Collections.Specialized.OrderedDictionary]::new()
				
				# Sort the hashtable by: 1) the length of the arrays in descending order and also 2) reverse the arrays for each key
				# WHY? #1 this makes sure the keys with greatest number of matches first
				#      #2 facilitates returning the most recent entry in each key
				# NOTE: Using Reverse method returns null. We need to use different methods for PS 5.1 (background and ISE), and PS7
				<# $PureOrQueryHt.GetEnumerator() | Sort-Object { $_.Value.Count } -Descending | ForEach-Object {
					$orderedHt.Add($_.Key, [array]::Reverse($_.Value.Clone()))  # Reverse the array for each key
				} #>
				if ($PSVersionTable.PSVersion.Major -lt 7) {
					# Reverse the array in place using the range operator
					$PureOrQueryHt.GetEnumerator() | Sort-Object { $_.Value.Count } -Descending | ForEach-Object {
						$orderedHt.Add($_.Key, $_.Value[($_.Value.Count -1 )..0])  
					}			
				} else {
					# Reverse the array in place using array slicing ([::]).
					# This method works by specifying a range from the last element (-1) to the first element (-($_.Value.Count))
					$PureOrQueryHt.GetEnumerator() | Sort-Object { $_.Value.Count } -Descending | ForEach-Object {
						$orderedHt.Add($_.Key, $_.Value[-1..-($_.Value.Count)])  # Reverse the array using array slicing
					}			
				}
				
				# For debugging
				# Write-Host ("$fn_name() orderedHt = " + ($orderedHt | ConvertTo-Json))

				# Expecting orderedHt to be like the following. 
				# ==> Compare it to the PureOrQueryHt example above:
				# ==> Note that key "not responding" has more entries, so it is first in order (at the top)
				# ==> Note that the array entries in BOTH keys are reverse sorted
				#	{
				#	  "not responding": [
				#	    "2\t10/20/2024 10:47:24.777 AM\tDW2\tAI: not responding",
				#	    "2\t10/20/2024 10:47:24.044 AM\tDW2\tAI: not responding",
				#	    "2\t10/20/2024 10:47:23.544 AM\tDW2\tAI: not responding",
				#	    "2\t10/20/2024 10:47:23.022 AM\tDW2\tAI: not responding",
				#	    "2\t10/20/2024 10:47:22.607 AM\tDW1\tAI: not responding"
				#	  ],
				#	  "fast enough": [
				#	    "2\t10/20/2024 12:47:24.777 PM\tBY\tClip: Disk can't write fast enough (0 1920 30010 0)",
				#	    "2\t8/22/2024 12:02:06.131 AM\tFD\tClip: Disk can't write fast enough (0 1918 30005 95802)",
				#	    "2\t8/22/2024 12:01:43.615 AM\tDW1\tClip: Disk can't write fast enough (0 2594 30010 74024)"
				#	  ]
				#	}
				
				# Initialize collector
				$lines = @() 
			
				$keys = $orderedHt.Keys

				# For debugging
				# Write-Host ("$fn_name() orderedHt keys_cnt = " + $keys.Count)
				# Write-Host ("$fn_name() orderedHt keys = " + ($keys | ConvertTo-Json))
				
				$break_cntr = 1
				do {
					foreach ($key in $keys) {
						if ($orderedHt["$key"]) {  # Just in case
							if ($orderedHt["$key"][0]) {
								# Add this line to the collector
								if (!($lines -contains $orderedHt[$key][0]) ){  # avoid duplicates
									$lines += $orderedHt[$key][0]
								}
							}
						}
					}
					$break_cntr++
				} until ($lines.Count -eq $max_rows -or $break_cntr -gt 5)
				
			}
		}

		# For debugging
		# Write-Host ("$fn_name() reversed entries = " + ($lines | ConvertTo-Json))

		# FORMAT EACH LINE
		$formatted_lines = @()  # Initialize accumulator
		$entry_no = 1
		foreach ($line in $lines) {

			# Split the logfile row/line by tabs
			$columns = $line -Split '\t'

			# For debugging
			# Write-Host "$fn_name() entry #$entry_no = $line"
			# Write-Host ("$fn_name() columns = " + ($columns | ConvertTo-Json))

			# Check if the row/line has at least 4 columns (to avoid index errors)
			if ($columns.Count -ge 4) {
				
				$level     = $columns[0].ToString().Trim()  # level is in the 1st column
				$timestamp = $columns[1].Trim()  # timestamp is in the 4th column
				$object    = $columns[2].Trim()  # object is in the 4th column
				$message   = $columns[3]         # message is in the 4th column
				
				# Apply highlighting to used levels
				if ($UsedLevels -contains $level) {
					$str = "<b>" + $entry_no + "</b>: <b>" + $level + "</b>, " + $timestamp + ", " + $object + ", " + $message
				} else {
					$str = "<b>" + $entry_no + "</b>: " + $level + ", " + $timestamp + ", " + $object + ", " + $message
				}
				
				# For debugging
				# Write-Host ("level = $level, type = " + $level.GetType().Name)
				# Write-Host "timestamp = $timestamp"
				# Write-Host "object = $object"
				# Write-Host "message = $message"
				# Write-Host "str = $str"
				
				$formatted_lines += $str
			}
			
			$entry_no++
			
			if ($entry_no -gt $max_rows) {
				break
			}
			
		}

		$rtn_str = $formatted_lines -Join ("<br>")

		# For debugging
		# Write-Host ("$fn_name() formatted_lines = " + ($formatted_lines | ConvertTo-Json))
		# Write-Host "$fn_name() rtn_str = $rtn_str"

		# APPLY HIGHLIGHTING (BOLD) TO USED TERMS
		# ... NOTE: we try to apply the highlighting so that the result preserves the original term's case 
		foreach ($term in $UsedTerms) {
			
			# Replace while preserving case
			# REMINDER: the -Replace operator is case insensitive, the .Replace() method is not
			# Use Regex class for case-insensitive replacement with case preservation
			# ... WORKS IN BOTH PS7 AND IN PS5 (Powershell ISE, Task Scheduler, ...)
			<# $rtn_str = [regex]::Replace($rtn_str, $term, { 
				param($match)
				if ($match.Value -cmatch "^[A-Z]") {
					# Uppercase match
					return "<b>$($match.Value.ToUpper())</b>"
				} elseif ($match.Value -cmatch "^[a-z]") {
					# Lowercase match
					return "<b>$($match.Value.ToLower())</b>"
				} else {
					# Mixed case match, return as-is
					return "<b>$($match.Value)</b>"
				}
			}, [System.Text.RegularExpressions.RegexOptions]::IgnoreCase)
			# For debugging
			Write-Host "rtn_str = $rtn_str" #>

			# Replace while preserving original capitalization (character-by-character)
			# ... WORKS IN BOTH PS7 AND IN PS5 (Powershell ISE, Task Scheduler, ...)
			# Convert term to a character array (this is required for this to work in PS5)
			# ... WHY? PS5 might return an error because it treats $term as a single object 
			#     rather than allowing direct character access like an array, e.g. $term[$i]
			$term_ary = $term.ToCharArray()  
			$rtn_str = [regex]::Replace($rtn_str, $term, { 
				param($match)			
				# Capitalize the replacement text based on the original match case
				$replacement = ""
				for ($i = 0; $i -lt $match.Value.Length; $i++) {
					if ($match.Value[$i] -cmatch "[A-Z]") {
						$replacement += $term_ary[$i].ToString().ToUpper()
					} else {
						$replacement += $term_ary[$i].ToString().ToLower()
					}
				}
				return "<b>$replacement</b>"

			}, [System.Text.RegularExpressions.RegexOptions]::IgnoreCase)
			# For debugging
			# Write-Host "rtn_str2 = $rtn_str"

		}
		
		return $rtn_str
	}

	function SendPushover() {
	# PURPOSE:  Send custom notification via the Pushover API
	# ARGS:     $MsgStr          = (string) notification message; may contain HTML
	#           $MsgTitle        = (string) notification title
	#           $PriorityLevel   = (integer) -2|-1|0|1|2; -2=lowest, -1=low(silent), 0=normal, 1=high-priority, 2=emergency
	#           $Sound           = (string) Pushover sound
	#           $Devices         = (string) Pushover device(s); comma-delimited
	#           $AttachPath      = (string) full filespec of resized alert image, or $null
	#           $Camera          = (string) Blue Iris camera shortname (replaces '&CAM' in user-setting $po_tags)
	#           $QuietFlg        = (boolean) $true prevents background messaging (showing the message sent)
	# RETURNS:  (string) Pushover API response, or 'ERR/' message
	# REQUIRES: none
	# GLOBALS:  Uses global vars: $script:po_app_token, $script:po_user_token, $script:po_sound, $script:po_device
	#           $script:po_retry, $script:po_expire, $script:po_tags,
	#           $script:FromConsole, $script:server_username
	#           $script:FromConsole = (boolean) $true notifies function that calling command was executed from the console
	#           $script:Quiet       = (boolean) $true suppresses verbose output to the console
	# ATTRIBUTIONS:
	#    handling attachments: https://tinyurl.com/mrxp7pn3 (PoshOver Powershell Package)
	#
	# V4 (2024-07-12)
	# V5 (2024-09-25) added -Sound & -Devices arguments
	# V6 (2024-10-15) added REQUIRES: and GLOBALS:
	# By jaydeel @ ipcamtalk.com
		
		param(
			[String]$MsgStr,      # [OPTIONAL] default is 'test'
			[String]$MsgTitle,    # [OPTIONAL] default is the App name associaled with your Pushover App token
			[Int]$PriorityLevel,  # [OPTIONAL] default is 0
			[String]$Sound,       # [OPTIONAL] default is ""
			[String]$Devices,     # [OPTIONAL] default is ""
			[String]$AttachPath,  # [OPTIONAL] default is null
			[String]$Camera,      # [OPTIONAL] default is ""
			[Boolean]$QuietFlg    # [OPTIONAL] default is $false
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered function: $fn_name()"
		# Write-Host "$fn_name() MsgStr = '$MsgStr'"
		# Write-Host "$fn_name() MsgTitle = '$MsgTitle'"
		# Write-Host "$fn_name() PriorityLevel = $PriorityLevel"
		# Write-Host "$fn_name() Sound = $Sound"
		# Write-Host "$fn_name() Devices = $Devices"
		# Write-Host "$fn_name() AttachPath = '$AttachPath'"	
		# Write-Host "$fn_name() Camera = '$Camera'"	
		# Write-Host "$fn_name() QuietFlg = $QuietFlg"
		# NOTE REGARDING MSG.EXE COMMANDS
		# ... You may need to replace 'msg.exe *' with 'msg.exe /server:computername username
		#msg.exe $script:server_username /TIME:20 "Pushover notification args...`npriority_level = $PriorityLevel`nQuietFlg = $QuietFlg`nAttachPath = $AttachPath`nMsgStr = $MsgStr"

		# HANDLE OPTIONAL COMMANDLINE ARGS
		if (!$MsgStr) {
			$MsgStr = "Alert w/ GIF attachment demo"
		}
		if (!$Camera) {
			$cam = ''
		} else {
			$cam = $Camera
		}
		if (!$Sound) {
			$Sound = $script:po_sound
		}
		if (!$Devices) {
			$Devices = $script:po_device
		}

		$rtn_str = $null
		
		#Prep the attachment
		if ($AttachPath) {
			#Get-Item -Path $AttachPath
			$attach_obj = (Get-Item -Path $AttachPath)
			$attach_bytes = [System.IO.File]::ReadAllBytes($AttachPath)
		}

		# For debugging
		#Write-Host ("$fn_name() attach_obj = " + $attach_obj)
		#Write-Host ("$fn_name() attach_obj.Name = " + $attach_obj.Name)
		#Write-Host ("$fn_name() po_tags = '" + $script:po_tags + "'")
		#Write-Host ("$fn_name() Camera = " + $Camera)
		
		# PREP CORE PARAMS
		# ... will be modified if sending an attachment
		$uri = 'https://api.pushover.net/1/messages.json'
		if ($script:po_tags) {
			$po_tags = ($script:po_tags).Replace("&CAM", $cam)
		} else {
			$po_tags = ''		
		}
		$parameters = @{
			token      = $script:po_app_token
			user       = $script:po_user_token
			priority   = $PriorityLevel
			retry      = $script:po_retry
			expire     = $script:po_expire
			tags       = $script:po_tags
			sound      = $Sound
			device     = $Devices
			html       = 1
			title      = $MsgTitle
			message    = $MsgStr
		}
		
		# For debugging
		#Write-Host ("parameters.Keys = " + $parameters.Keys)
		#Write-Host "`nPushover parameters..."
		<# [string[]]$arr = $parameters | Out-String -stream
		foreach ($item in $arr) {
			if ($item) {
				Write-Host $item
			}
		} #>
		
		# EXECUTE
		if (!$AttachPath -or ($attach_bytes.Length -eq 0)) {
			
			# SEND BASIC (TEXT-ONLY) NOTIFICATION

			# ... NOTE: When running in the background, both Invoke-RestMethod & Invoke-WebRequest
			#	  do not execute; it doesn't produce anything, no error, warning, or output.
			#     REF: https://stackoverflow.com/a/50274056 looked like a possible solution
			#     however, since PS v6.0 switch -UseBasicParsing has been deprecated, and 
			#     all web requests now use basic parsing only.
			# ... attachmnets
			#     https://psg-prod-centralus.cloudapp.net/packages/Poshover/0.1.450600-alpha/Content/Private%5CSend-MessageWithAttachment.ps1
			try {
				$ApiResponse = $parameters | Invoke-RestMethod -Uri $uri -Method Post
				# Alternative syntaxes
				#$parameters | Invoke-RestMethod -Uri $uri -Method Post | Out-Null
				#$ApiResponse = Invoke-RestMethod -Uri $uri -Method Post -Form $parameters
			}
			catch {
				# For debugging
				# Write-Warning "An error was encountered sending the Pushover text message:"
				# Write-Error $_
				
				$rtn_str = "ERR/an error was was encountered sending the Pushover text message"
				return $rtn_str
			}
				
			# Expecting string like "@{status=1; request=21d4f100-3127-4c62-aad9-27058400fd74}"
			[String]$response_str = ($ApiResponse | ConvertTo-Json -Depth 100) | ConvertFrom-Json
					
			# For debugging
			#Write-Host "$fn_name() w/o attachment ApiResponse = $response_str"

		} else {	
		
			# SEND NOTIFICATION WITH ATTACHMENT
			
			# Notes:
			# Modified code from:  https://tinyurl.com/mrxp7pn3
			# More refs: https://gist.github.com/markekraus/a0135d0a5f52b609eda513df4db71cab

			try {

				Add-Type -AssemblyName System.Net.Http
				# NOTE: Originally, the $client= line below was not working when this script was 
				#        executed in the background via a Blue Iris 'run a program' action
				#        SOLUTION:  https://stackoverflow.com/a/70535083
				$client = [system.net.http.httpclient]::new()

				try {
					$content = [system.net.http.multipartformdatacontent]::new()
					foreach ($key in $parameters.Keys) {
						$textContent = [system.net.http.stringcontent]::new($parameters.$key)
						$content.Add($textContent, $key)
					}
					
					# For debugging
					#Write-Host ("content = " + $content | ConvertTo-Json)
					
					$jpegContent = [system.net.http.bytearraycontent]::new($attach_bytes)
					$jpegContent.Headers.ContentType = [system.net.http.headers.mediatypeheadervalue]::new('image/jpeg')
					$jpegContent.Headers.ContentDisposition = [system.net.http.headers.contentdispositionheadervalue]::new('form-data')
					$jpegContent.Headers.ContentDisposition.Name = 'attachment'
					$jpegContent.Headers.ContentDisposition.FileName = $AttachPath
					$content.Add($jpegContent)

					#For debugging
					# Write-Host ("content = " + ($content | ConvertTo-Json))

					# Not needed.
					# REF: https://stackoverflow.com/questions/43014894/httpclient-and-stream-issues
					#$content.LoadIntoBufferAsync().GetAwaiter().GetResult()

					# Modify the message body
					Write-Verbose "Message body:`r`n$($content.ReadAsStringAsync().Result.Substring(0, 2000).Replace($parameters.token, "********").Replace($parameters.user, "********"))"

					# For debugging
					#msg.exe $script:server_username /TIME:10 "executing pushover api cmd w/ attachment..."

					# DO IT!
					$ApiResponse = $client.PostAsync($uri, $content).Result
					
					# For debugging
					#Write-Host ("$fn_name() w/ attachment ApiResponse = " + ($ApiResponse | ConvertTo-Json -Depth 32))
					
					try {
						$response_str = $ApiResponse.Content.ReadAsStringAsync().Result | ConvertFrom-Json
						
						# For debugging
						#Write-Host "$fn_name() ApiResponse.Content = $response_str"
					}
					catch {
						$response_str = "WARNING: Unexpected Pushover API response; attachment may be too large"
					}
					
					# For debugging
					#Write-Host "$fn_name() w attachment ApiResponse = $response_str"
					
				}
				catch {
					$response_str = "ERR/unable to send Pushover notification"				
				}
				finally {
					$content.Dispose()  # Release resources
				}
			}
			finally {
				$client.Dispose()  # Release resources
			}
		}

		# For testing - Simulate a bad pushover response
		#$response_str = 'invalid'
		#$response_str = 'status=0'
		
		if ($response_str -match 'invalid') {
			$rtn_str = "ERR/Pushover API returned 'invalid'"
		} elseif (!($response_str -match 'status=1')) {
			$rtn_str = "ERR/failed to send Pushover notification"				
		} else {
			# SUCCESS!!
			$rtn_str = "" + $response_str
		}
		
		return $rtn_str
	}

	function SendTelegram() {
	# PURPOSE: Send custom notification via the Telegram Bot API
	# ARGS:    $MsgStr          = (String) notification message; may contain HTML
	#          $PriorityLevel   = CURRENTLY UNSUPPORTED
	#          $AttachPath      = (String) full filespec of attachment, or $null
	#          $Camera          = (String) Blue Iris camera shortname (replaces '&CAM' in user-setting $po_tags)
	# RETURNS: (String) Telegram API response, or 'ERR/' message
	# GLOBALS: $script:FromConsole, $script:Quiet
	# ATTRIBUTIONS:
	#    handling attachments: https://tinyurl.com/mrxp7pn3 (PoshOver Powershell Package)
	#
	# NOTES: 
	# 1. Per the API, 'sendPhoto' max is 10 MB  https://core.telegram.org/bots/api#sendPhoto
	# 2. Per the API, 'sendVideo' max is 50 MB  https://core.telegram.org/bots/api#sendVideo
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# V1 (7/12/2024)
	# V2 (9/23/2024) fixed bug where $response_str could be type PSCutomObj
	# V3 (10/4/2024) fixed bug where $MsgStr cannot contain HTML '<br>' tags
	# V4 (11/12/2024) fixed bug where $response_str was actually a PSCustomObject; 
	#                 fixed bug where Telegram's sendVideo method stopped working for GIFs; changed to sendAnimation method
	# V5 (11/22/2024) removed arguments -FromConsoleFlg, -QuietFlg - now using globals; misc tweaks
		
		param(
			[String]$MsgStr,      # [OPTIONAL] default is 'test'
			[String]$AttachPath,  # [OPTIONAL] default is null
			[String]$Camera       # [OPTIONAL] default is ""
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered function: $fn_name()"
		# Write-Host "$fn_name() MsgStr = '$MsgStr'"
		# Write-Host "$fn_name() AttachPath = '$AttachPath'"	
		# Write-Host "$fn_name() Camera = '$Camera'"	
		# NOTE REGARDING MSG.EXE COMMANDS
		# ... You may need to replace 'msg.exe *' with 'msg.exe /server:computername username
		#msg.exe $script:server_username /TIME:20 "Pushover notification args...``nAttachPath = $AttachPath`nMsgStr = $MsgStr"

		# HANDLE OPTIONAL COMMANDLINE ARGS
		if (!$MsgStr) {
			$MsgStr = "test"
		} else {
			# The Telegram API does not support '<br>' tags
			$MsgStr = $MsgStr.Replace("<br>", "`n")
		}
		
		if (!$Camera) {
			$cam = ''
		} else {
			$cam = $Camera
		}

		$PsVersion = $PSVersionTable.PSVersion.Major

		$rtn_str = $null
		
		#Prep the attachment
		if ($AttachPath) {
			try {
				$attach_obj = Get-Item -Path $AttachPath -ErrorAction Stop
			}
			catch {
				if ($script:FromConsole) {
					Write-Warning "The specified attachment file could not be interpreted properly"
				}
				$rtn_str = $false
				return $rtn_flg
			}
			$attach_bytes = [System.IO.File]::ReadAllBytes($AttachPath)
		}

		# For debugging
		# Write-Host ("$fn_name() attach_obj = " + $attach_obj)
		# Write-Host ("$fn_name() attach_obj.Name = " + $attach_obj.Name)
		# Write-Host ("$fn_name() cam = " + $cam)
		
		# PREP CORE PARAMS
		# ... will be modified if sending an attachment
		$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendMessage"

		$parameters = @{
			chat_id     = $script:tg_chat_id
			parse_mode  = 'html'
		}
		
		# EXECUTE
		if (!$AttachPath -or ($attach_bytes.Length -eq 0)) {

			# SEND BASIC NOTIFICATION (NO ATTACHMENT)
			
			$parameters.Add('text', $MsgStr)
		
			# For debugging
			#Write-Host ("parameters.Keys = " + $parameters.Keys)
			<# Write-Host "`nPushover parameters..."
			[string[]]$arr = $parameters | Out-String -stream
			foreach ($item in $arr) {
				if ($item) {
					Write-Host $item
				}
			} #>

			# ... NOTE: When running in the background, both Invoke-RestMethod & Invoke-WebRequest
			#	  do not execute; it doesn't produce anything, no error, warning, or output.
			#     REF: https://stackoverflow.com/a/50274056 looked like a possible solution
			#     however, since PS v6.0 switch -UseBasicParsing has been deprecated, and 
			#     all web requests now use basic parsing only.

			try {
				if ($PsVersion -gt 5) {
					# Returns a PSCustomObject
					$api_response_obj = `
						Invoke-RestMethod `
						-Uri $uri `
						-Method Post `
						-Form $parameters `
						-ErrorAction Stop
				} else {
					# Returns a PSCustomObject
					$api_response_obj = `
						Invoke-RestMethod `
						-Uri $uri `
						-Method Post `
						-Body $parameters `
						-ErrorAction Stop
				}
			}
			catch {
				if ($script:FromConsole) {
					Write-Warning "An error was encountered sending the Telegram text message:"
					Write-Error $_
				}
				$rtn_str = "ERR/an error was was encountered sending the Telegram text message"
				return $rtn_str
			}
			
			# Expecting string like "@{ok=True; result=}"
			$response_str = [string](($api_response_obj | ConvertTo-Json -Depth 100) | ConvertFrom-Json)
					
			# For debugging
			#Write-Host "$fn_name() w/o attachment api_response_obj = $response_str"

		} else {
			
			# SEND NOTIFICATION WITH ATTACHMENT
			
			# Notes:
			# Modified code from:  https://www.powershellgallery.com/packages/PoshGram/0.8.4/Content/PoshGram.psm1

			try {
				$fileObject = Get-Item $AttachPath -ErrorAction Stop
			}
			catch {
				if ($script:FromConsole) {
					Write-Warning "The specified attachment file could not be interpreted properly"
				}
				$rtn_str = "ERR/the specified attachment file could not be interpreted properly"
				return $rtn_str
			}

			# For debugging
			# Write-Host ("$fn_name() fileObject = " + ($fileObject | ConvertTo-Json))

			if ($PsVersion -ge 7) {

				if ($AttachPath.EndsWith('.jpg')) {
					$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendPhoto"
					$parameters.Add('photo', $fileObject)
					$parameters.Add('caption', $MsgStr)
				} elseif ($AttachPath.EndsWith('.gif') -or $AttachPath.EndsWith('.mp4')) {
					#$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendVideo"
					#$parameters.Add('video', $fileObject)
					$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendAnimation"
					$parameters.Add('animation', $fileObject)
					$parameters.Add('caption', $MsgStr)			
				} else {
					if ($script:FromConsole) {
						Write-Warning "Unsupported attachment file (expecting *.jpg, *.gif, or *.mp4)"
					}
					$rtn_str = "ERR/unsupported attachment file (expecting *.jpg, *.gif, or *.mp4)"
					return $rtn_str			
				}
		
				# For debugging
				# Write-Host "$fn_name() uri = $uri"
				# Write-Host ("parameters.Keys = " + $parameters.Keys)
				# Write-Host "`nTelegram parameters..."
				<# [string[]]$arr = $parameters | Out-String -stream
				foreach ($item in $arr) {
					if ($item) {
						Write-Host $item
					}
				} #>
					
				try {
					# Returns a PSCustomObject
					$api_response_obj = `
						Invoke-RestMethod `
						-Uri $uri `
						-Method Post `
						-Form $parameters `
						-ErrorAction Stop				
				}
				catch {
					if ($script:FromConsole) {
						Write-Warning "An error was encountered sending the Telegram video message:"
						Write-Error $_
					}
					$rtn_str = "ERR/an error was encountered sending the Telegram video message"
					return $rtn_str			
				}
					
				# Expecting string like "@{ok=True; result=}"
				[String]$response_str = ($api_response_obj | ConvertTo-Json -Depth 100) | ConvertFrom-Json
				# For debugging
				Write-Host ("$fn_name() api_response_obj type = " + $api_response_obj.GetType())
				# Write-Host ("$fn_name() api_response_obj = " + ($api_response_obj | ConvertTo-Json -Depth 100))
				# Write-Host ("$fn_name() api_response_obj.ok = " + $api_response_obj.ok)
				# Write-Host ("$fn_name() response_str type = " + $response_str.GetType())
				# Write-Host ("$fn_name() response_str = " + $response_str)
					
			} else {

				if ($AttachPath.EndsWith('.jpg')) {
					$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendPhoto"
					$parameters.Add('caption', $MsgStr)
				} elseif ($AttachPath.EndsWith('.gif') -or $AttachPath.EndsWith('.mp4')) {
					$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendVideo"
					$parameters.Add('caption', $MsgStr)			
				} else {
					if ($script:FromConsole) {
						Write-Warning "Unsupported attachment file (expecting *.jpg, *.gif, or *.mp4)"
					}
					$rtn_str = "ERR/unsupported attachment file (expecting *.jpg, *.gif, or *.mp4)"
					return $rtn_str			
				}
		
				# For debugging
				# Write-Host "$fn_name() uri = $uri"
				# Write-Host ("parameters.Keys = " + $parameters.Keys)
				# Write-Host "`nTelegram parameters..."
				<# [string[]]$arr = $parameters | Out-String -stream
				foreach ($item in $arr) {
					if ($item) {
						Write-Host $item
					}
				} #>
				
				try {
					
					# This command fails in PS5
					#$rtn_flg = Invoke-RestMethod -Uri $uri -Method Post -Body $parameters -ErrorAction Stop

					#Prep the attachment
					if ($AttachPath) {
						#Get-Item -Path $AttachPath
						$attach_obj = Get-Item -Path $AttachPath
						$attach_bytes = [System.IO.File]::ReadAllBytes($AttachPath)
					}

					Add-Type -AssemblyName System.Net.Http
					# NOTE: Originally, the $client= line below was not working when this script was 
					#        executed in the background via a Blue Iris 'run a program' action
					#        SOLUTION:  https://stackoverflow.com/a/70535083
					$client = [system.net.http.httpclient]::new()

					try {
						$content = [system.net.http.multipartformdatacontent]::new()
						foreach ($key in $parameters.Keys) {
							$textContent = [system.net.http.stringcontent]::new($parameters.$key)
							$content.Add($textContent, $key)
						}
						
						# For debugging
						# Write-Host ("content = " + $content | ConvertTo-Json)
						
						$videoContent = [system.net.http.bytearraycontent]::new($attach_bytes)
						$videoContent.Headers.ContentType = [system.net.http.headers.mediatypeheadervalue]::new('image/jpeg')
						$videoContent.Headers.ContentDisposition = [system.net.http.headers.contentdispositionheadervalue]::new('form-data')
						$videoContent.Headers.ContentDisposition.Name = 'video'  # Telegram API 'video' parameter
						$videoContent.Headers.ContentDisposition.FileName = $AttachPath
						$content.Add($videoContent)

						#For debugging
						# Write-Host ("content = " + ($content | ConvertTo-Json))

						# Not needed.
						# REF: https://stackoverflow.com/questions/43014894/httpclient-and-stream-issues
						#$content.LoadIntoBufferAsync().GetAwaiter().GetResult()

						# Modify the message body
						Write-Verbose "Message body:`r`n$($content.ReadAsStringAsync().Result.Substring(0, 2000).Replace($parameters.chat_id, "********"))"

						# For debugging
						#msg.exe $script:server_username /TIME:10 "executing telegram api cmd w/ 'video' parameter..."

						# DO IT!
						$api_response_obj = $client.PostAsync($uri, $content).Result
						
						# For debugging
						# Write-Host ("api_response_obj w/ 'video' parameter = " + ($api_response_obj | ConvertTo-Json -Depth 100))
						
						try {
							$response_str = $api_response_obj.Content.ReadAsStringAsync().Result | ConvertFrom-Json
							
							# For debugging
							# Write-Host "api_response_obj.Content = $response_str"
						}
						catch {
							$response_str = "WARNING: Unexpected Telegram API response; 'video' parameter may be too large"
						}
						
						# For debugging
						# Write-Host "api_response_obj w/ 'video' parameter = $response_str"
						
					}
					catch {
						$response_str = "ERR/unable to send Telegram notification"				
					}
					finally {
						$content.Dispose()  # Release resources
					}
				}
				finally {
					$client.Dispose()  # Release resources
				}
			}
			
		}

		# For testing - Simulate a bad Telegram response
		#$response_str = 'ok=False'
		
		if (!($response_str -match "ok=True")) {
			if ($script:FromConsole) {
				Write-Warning "failed to send Telegram notification"
			}
			$rtn_str = "ERR/failed to send Telegram notification"				
			$rtn_flg = $false
			# Write-Error $_
		} else {
			# SUCCESS!!
			$rtn_str = $response_str
		}
		
		return $rtn_str
	}

	function LogScriptResultsToCsv() {
	# PURPOSE:  Log performance data to CSV file.
	# ARGS:     see param lines below
	# RETURNS:  (string) 'success' or 'ERR/unable to write to CSV log'
	# NOTES:    1. CSV file is created automatically if it does not already exist. 
	#           2. CSV file's header row is generated from the object keys in variable $new_content_obj.

		param(
			[String]$CsvFilespec,
			[String]$Timestamp,
			[Int]$NewestRowsCnt,
			[Int]$MatchedRowsCnt,
			[String]$PushSent,
			[String]$ScriptMsec,
			[String]$ScriptHost,
			[String]$ArgsUsedStr
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered function: $fn_name()"
		# Write-Host "$fn_name() CsvFilespec = $CsvFilespec"
		# Write-Host "$fn_name() Timestamp = $Timestamp"
		# Write-Host "$fn_name() argsused = '$argsused'"

		# HANDLE REQUIRED COMMANDLINE ARGS
		if (!$CsvFilespec) {
			$err_msg = "ERR/expecting 2 arguments (-CsvFilespec -Timestamp)"
			return $err_msg
		}
		if (!$Timestamp) {
			$err_msg = "ERR/expecting 2 arguments (-CsvFilespec -Timestamp)"
			return $err_msg
		}
		
		$rtn_str = 'success'
			
		# BUILD THE NEW ROW
		# ... Note: this object's keys are the header column labels
		$new_content_obj = [PSCustomObject]@{
			'Timestamp'      = $Timestamp
			'LogfileNewRows' = $NewestRowsCnt
			'MatchedRows'    = $MatchedRowsCnt
			'PushSent'       = $PushSent
			'ScrMsec'        = $ScriptMsec
			'ParentProcess'  = $ScriptHost
			'ArgsUsed'       = $ArgsUsedStr
		}
		
		# APPEND NEW ROW TO CSV FILE, creating the file if necessary
		if ($PSVersionTable.PSVersion.Major -ge 7) {
			# REFS: https://tinyurl.com/yvy4mjy6
			#       https://stackoverflow.com/a/70717331
			#       https://stackoverflow.com/a/60681762
			try {	
				# NOTE: To simulate an Export-Csv error, remove the -Force argument and make file read-only
				$new_content_obj | Export-Csv -LiteralPath $CsvFilespec -Append -UseQuotes AsNeeded
				#$new_content_obj | Export-Csv -LiteralPath $CsvFilespec -Append -Force -UseQuotes AsNeeded
			} catch {
				$err_msg = "ERR/unable to append row to CSV file; $_"
				return $err_msg
			}
		} else {
			try {
				# NOTE: To simulate an Export-Csv error, remove the -Force argument and make file read-only
				$new_content_obj | Export-Csv -LiteralPath $CsvFilespec -Append -Force -NoTypeInformation
			} catch {
				$err_msg = "ERR/unable to append row to CSV file; $_"
				return $err_msg
			}
		}

		if (!(Test-Path -LiteralPath $CsvFilespec)) {
			$err_msg = "ERR/unable to write to '$CsvFilespec'"
			return $err_msg
		}
		
		return $rtn_str
	}

	function PostCustomBiLogEntry() {
	# PURPOSE:  Add custom message to the Blue Iris logfile
	# ARGS:     -MsgStr    = (string) NOTE: be sure spaces are replaced with '%20'
	#           -Level     = (integer) 0|1|2 (0:info, 1:warn, 2:error)
	#                         NOTE: level=2 can trigger a status alert if the 'Send only if there are new Errors' box 
	#                         is checked on the 'Status alerts' dialog (see button at bottom of the 'Blue Iris Status' (log) dialog)
	#           -SilentFlg = (boolean) $true prevents background-only msg.exe pop-up (showing the message sent)
	# RETURNS:  (string) message as posted to the Blue Iris logfile. 
	#                   OR ERR message
	# REQUIRES:  none
	# GLOBALS:   Uses vars: $script:bi_webserver, $script:post_w_creds_flg, $script:bi_user_name, $script:bi_user_pw
	# KEY POINT: the message string maximum length is 127 characters (BI HTTP interface command '/admin?log=message&level=x')
	#
	# V2 (9/30/2024) fixed illegal URL chars in $MsgStr; tweaked global var usage
	# V3 (10/12/2024) trimmed message to not exceed the Blue Iris API's maximum length of 127 characters
	# by jaydeel at ipcamtalk.com

		param (
			[String]$MsgStr,
			[Int]$Level,
			[Boolean]$SilentFlg
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
		
		# For debugging
		# Write-Host "Entered function: $fn_name()"
		# Write-Host ("$fn_name() MsgStr len = " + $MsgStr.Length)
		# Write-Host "$fn_name() MsgStr = '$MsgStr'"
		# Write-Host "$fn_name() Level = '$Level'"
		# Write-Host "$fn_name() SilentFlg = '$SilentFlg'"
		
		
		if (!$script:bi_webserver) {
			$err_msg = "ERR/bad web server address"
			return $err_msg
		}
		
		# Blue Iris webserver address:port
		$http = "http://" + $script:bi_webserver
		
		# For testing
		# $http += "1"
		
		# Fix unallowed characters for a URL
		# ... Note: that 2nd line is preferred b/c it supports Regex, and is regarded to be more readable
		#     The 1st line is marginally faster, but this only noticeable for very large/complex strings
		#$msg_str = $MsgStr.Replace("`n", "%20").Replace("\\r",  "%20").Replace("\\n", "%20").Replace(" ", "%20")
		$msg_str = $MsgStr.Replace("`n", "%20").Replace("\\r",  "%20").Replace("\\n", "%20").Replace(" ", "%20")
		
		# For debugging
		# Write-Host ("$fn_name() msg_str BEFORE len = " + $msg_str.Length)
		# Write-Host "$fn_name() msg_str BEFORE = '$msg_str'"

		# For testing
		# ... if used, the Blue Iris logfile entry will drop the final 'g'
		#$msg_str = "thisisalongmessagethatisonehundredandtwentyeightcharacterslong1234thisisalongmessagethatisonehundredandtwentyeightcharacterslong"

		# Trim message if it exceeds Blue Iris maximum length (127 characters)
		# ... and append an ellipsis (3 dots "...", not the character "…", which displays as 'â€¦')
		if ($msg_str.Length -ge 127) {
			$msg_str = $msg_str.Substring(0, 124) + "..."
		} 
		
		# For debugging
		# Write-Host ("$fn_name() msg_str AFTER len = " + $msg_str.Length)
		# Write-Host "$fn_name() msg_str AFTER = '$msg_str'"

		if ($script:post_w_creds_flg) {
			$uri = "$http/admin?log=$msg_str&level=$Level&user=$script:bi_username&pw=$script:bi_pw"
		} else {
			$uri = "$http/admin?log=$msg_str&level=$Level"
		}
		
		# For debugging
		# Write-Host "$fn_name() uri = $uri"
		
		# DO IT!
		# NOTE: this command will NOT fail just because the URL $msg_str exceeds 127 characters
		try {
			$web_response_obj = `
				Invoke-WebRequest `
				-Uri $uri
		} 
		catch {

			# For debugging
			# Write-Host $error
			# Write-Host $_.ScriptStackTrace
			# Write-Host $PSitem

			# For debugging
			# Write-Host ("$fn_name() " + $_.Exception.Message) -ForegroundColor Red
			
			$err_msg = "ERR/failure sending HTTP Interface command to the Blue Iris server:`n" + $_.Exception.Message
			return $err_msg
		}
		
		# For debugging
		# Write-Host ("$fn_name() web_response_obj =`n" + $web_response_obj)	
		
		# For testing
		#$web_response_obj = "xxx"
		
		# Expecting response_obj like,  
		#	signal=green
		#	schedule=Home
		#	profile=1
		#	lock=0
		#	transport=0,0/1
		#	dio_inputs=0
		
		if ($web_response_obj -match 'signal=' -and $web_response_obj -match 'schedule=' -and $web_response_obj -match 'profile=') {
			# Normal response
		} else {
			# Unexpected response
			$err_msg = "ERR/unexpected HTTP Interface command response from Blue Iris server:`n$web_response_obj"
			return $err_msg
		}

		# Restore spaces to the message
		$rtn_str = $msg_str.Replace("%20"," ")

		# HANDLE NOTIFICATIONS
		if (!$SilentFlg) {
			if (!$script:FromConsole -and $script:bg_msgs_on) {
				msg.exe * /TIME:10 "$fn_name() Custom BI logfile message posted...`nlevel = $level`n$rtn_str"
			}
		}
		
		<# if ($rtn_str.StartsWith('ERR/')) {
			$rtn_str = "Custom BI logfile message posted"
		} else {
			$rtn_str = "Custom BI logfile message posted:`n'" + ($MsgStr.Replace("%20"," ")) + "'`n"
		} #>

		return $rtn_str
	}

	function GetLogfileLineCount() {
	# PURPOSE: efficiently return the number of lines in large text file
	# via ChatGPT... For the greater efficiency, especially for very large files, use .NET’s StreamReader
	# This method processes one line at a time without loading the entire file into memory.
	#
		param (
			[String]$FilePath
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered function: $fn_name()"
		# Write-Host "$fn_name() FilePath = '$FilePath'"
		
		# $timer = [Diagnostics.Stopwatch]::StartNew()

		$lineCount = 0
		$reader = [System.IO.StreamReader]::new("$FilePath")
		while ($reader.ReadLine() -ne $null) {
			$lineCount++
		}
		$reader.Close()

		# $timer.stop()
		# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
		# $elapsed_msec_str = $elapsed_msec.ToString("0.00")
		# Write-Host ("GetLogfileLineCount() timer = $elapsed_msec_str msec")

		# For debugging
		# Write-Host "$fn_name() lineCount = '$lineCount'"

		return $lineCount
		
	}

	function PauseOutput($message) {
	# PURPOSE:   Pause lengthy output before exceeding console height3
	# KEY-POINT: Works only 
	# SOURCE:    https://stackoverflow.com/a/28237896/13499984

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host "$fn_name() -message = '$message'"

		# Check if running Powershell ISE
		$isPsISE = $Host.Name -match 'ISE Host'
		#Write-Host ("HostName = " + $Host.Name)
		# Write-Host "isISE = $isPsISE"

		if ($isPsISE) {
			Add-Type -AssemblyName System.Windows.Forms
			[System.Windows.Forms.MessageBox]::Show("$message")
		} else {
			Write-Host "$message" -ForegroundColor Yellow
			$x = $host.ui.RawUI.ReadKey("NoEcho,IncludeKeyDown")
		}
	}

	function KillRecursiveMgtFiles() {
		
		param(
			[String]$RootPath
		)

		# Remove previously used management files
		if (Test-Path -Path "$RootPath\*" -Include "batchfile_arg_*.txt") {
			Remove-Item -Path "$RootPath/batchfile_arg*.txt"
		}
		
	}

	function PauseOutput($message) {
	# PURPOSE:   Pause lengthy output before exceeding console height3
	# KEY-POINT: Works only 
	# SOURCE:    https://stackoverflow.com/a/28237896/13499984

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host "$fn_name() -message = '$message'"

		# Check if running Powershell ISE
		$isPsISE = $Host.Name -match 'ISE Host'
		#Write-Host ("HostName = " + $Host.Name)
		# Write-Host "isISE = $isPsISE"

		if ($isPsISE) {
			Add-Type -AssemblyName System.Windows.Forms
			[System.Windows.Forms.MessageBox]::Show("$message")
		} else {
			Write-Host "$message" -ForegroundColor Yellow
			$x = $host.ui.RawUI.ReadKey("NoEcho,IncludeKeyDown")
		}
	}

	function Format-Color() {
	# PURPOSE:  Colorize the command's output based on line-by-line pattern matching.
	#           Also supports pausing output based on console's line height
	# SOURCE:   https://www.bgreco.net/powershell/format-color/
	# REFS:     https://superuser.com/a/1709091  (colorize output)
	#           https://stackoverflow.com/a/4647985/13499984  (colorize output)
	#           https://stackoverflow.com/a/44872079/13499984  (measure window height)
	# REQUIRES: PauseOutput()

		param (
			[Hashtable]$InputObject = @{}
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
		
		$Colors      = $InputObject.Colors
		$SimpleMatch = $InputObject.SimpleMatch
		$PauseOutput = $InputObject.PauseOutput

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host ("$fn_name() -InputObject =" + ($InputObject | ConvertTo-Json))
		# Write-Host ("Colors type = " + $Colors.GetType().Name)
		# Write-Host ("Colors = " + ($Colors | ConvertTo-Json))
		# Write-Host "SimpleMatch = $SimpleMatch"
		# Write-Host "PauseOutput = $PauseOutput"
		# Write-Host ("console_lines = " + $Host.UI.RawUI.WindowSize.Height)

		$output_lines = ($input | Out-String) -replace "`r", "" -split "`n"
		$i = 0
		foreach($line in $output_lines) {
			$color = ''
			foreach($pattern in $Colors.Keys){
				if(!$SimpleMatch -and $line -match $pattern) { 
					$color = $Colors[$pattern] 
				} elseif ($SimpleMatch -and $line.IndexOf($pattern) -gt -1) { 
					$color = $Colors[$pattern]
				}
			}

			if($color) {
				Write-Host $line -ForegroundColor $color 			
			} else {
				Write-Host $line
			}

			if ($PauseOutput) {
				if ($Host.UI.SupportsVirtualTerminal) {
					# KEY-PONT: This feature supports only Consoles.
					# The Powershell ISE UI does not have a height because it is a Forms window. Its viewport 
					# has a size, but it is not tied to the number of visible lines
					$console_lines = $Host.UI.RawUI.WindowSize.Height
					if ($i -eq $console_lines - 5) {
						PauseOutput('Press any key to continue...')
					}
				}
			}

			$i++
		}
	}

}

# GET BLUE IRIS SETTINGS FROM THE WINDOWS REGISTRY
# ... KEY POINT: this section must precede the -Examples argument output
if (1 -eq 1) {  # Use to collapse this code section while editing

	# GET BLUE IRIS LOG FILE PATH
	$bi_logfiles_path = GetBiLogPath
	
	# For testing
	# $bi_logfiles_path += "1"

	# For debugging
	#Write-Host "bi_logfiles_path = '$bi_logfiles_path'"

	if (!$bi_logfiles_path) {
		$err_str = "ERR/unable to get the Blue Iris logfile path from the registry"
		return $err_Str
	}
	if (!(Test-Path -LiteralPath $bi_logfiles_path)) {
		$err_str = "ERR/Blue Iris logfile path from the registry does not exist: '$bi_logfiles_path'"
		return $err_Str
	}

	# GET BLUE IRIS WEB SERVER LAN IP AND PORT
	# ... failure will NOT terminate script; however logging to the Blue Iris logfile will be disabled
	$bi_webserver = GetBiWebserverIpPort
	
	# For testing
	#$bi_webserver += "1"

	# For debugging
	# Write-Host "bi_webserver = $bi_webserver"
	
	if (!$bi_webserver) {
		if ($FromConsole) {
			Write-Warning "unable to get Blue Iris web server address:port from the registry"
		}
	}
	
	# @@@ DEV GET BLUE IRIS WEB SERVER USERNAME'S PASSWORD
	<# $bi_user_pw_raw = GetBiWebserverUserPw -UserName $bi_user_name
	if (!$bi_user_pw_raw) {
		$err_str = "Sorry - unable to get the registry password for Blue Iris user '$bi_user_name'"
		return $err_Str
	} elseif ($bi_user_pw_raw.StartsWith("ERR/")) {
		$err_str = "Sorry - encountered an error trying to get the registry password for Blue Iris user '$bi_user_name'"
		return $err_Str
	}
	Write-Host "bi_user_pw_raw = '$bi_user_pw_raw'" #>
}

# SHOW HELP
if ($PSBoundParameters.ContainsKey('Help') -and $FromConsole) {
	Write-Host "============================================================================================"
	Write-Host "BLUE IRIS LOGFILE WATCHDOG UTILITY"
	Write-Host "Inspect recent entries in the active BI logfile for specific column values;"
	Write-Host "If matches are found, send a notification via Pushover or Telegram (or both)"
	Write-Host "--------------------------------------------------------------------------------------------"
	Write-Host "Usage: $ScriptName -Examples"
	Write-Host "Args:"
	Write-Host "   -FilterLevels  ... terms to find in the 'level' column; comma-delimited [1,2,3]"
	Write-Host "   -FilterObjects ... terms to find in the 'object' column; comma-delimited [2,3,4]"
	Write-Host "   -FilterMsgs    ... terms to find in the 'message' column; comma-delimited [2,3,4,5]"
	Write-Host "   -BatchFile ... use batch mode [6,7]; include full path if file not in `$app_path [8]"
	Write-Host "   -Reset     ... clear temp files used to recursively execute script via -BatchFile"
	Write-Host "   -Renew     ... clear temp files used to track number of new rows added to the logfile"
	Write-Host "   -ShowMatches ... list the logfile entries that match the filters"
	Write-Host "   -PushService ... override default push service ('Pushover', 'Telegram', or 'all') [8]"
	Write-Host "   -Title       ... custom title for notification"
	Write-Host "   -Priority    ... custom priority for notification (Pushover only)"
	Write-Host "   -Sound       ... custom sound for notification (Pushover only)"
	Write-Host "   -Devices     ... custom devices for notification (Pushover only)"
	Write-Host "   -Help     ... show this usage guide"
	Write-Host "   -Examples ... show the examples guide"
	Write-Host "   Useful for testing:"
	Write-Host "   -TestFile ... use *.txt file instead of the active logfile [9]"
	Write-Host "   -NoPush   ... do not send notification"
	Write-Host "   -NoBiLog  ... disable logging to the Blue Iris logfile"
	Write-Host "   -DevMode  ... show developer mode info"
	Write-Host "Notes: "
	Write-Host "[1] Expected values: 1=informational, 2=warning, or 3=error"
	Write-Host "[2] Values for these arguments must be like val1,val2, not val1,val2"
	Write-Host "[3] Values for these arguments may use the NOT logical operator '!val1'"
	Write-Host "[4] Values for these arguments must be quoted if they contain spaces"
	Write-Host "[5] Values for these arguments must use %2C for literal commas"
	Write-Host "[6] Like -BatchFile 'my_filters.txt' (OR 'D:\path\my_filters.txt')"
	Write-Host "[7] Batch mode's *.txt file should have content like the following:"
	Write-Host "       # comment"
	Write-Host '       -FilterLevels=1|2'
	Write-Host "       # comment"
	Write-Host '       -FilterLevels=2;-FilterMsgs=fail,corrupt'
	Write-Host "       # comment"
	Write-Host '       -FilterLevels=2;-FilterObjects=DB|MQTT;-FilterMsgs=fail|corrupt'
	Write-Host "[8] See the user-settings-file"
	Write-Host ("[9] Content must have 4 tab-delimited entries per line (level") -NoNewLine
	Write-Host ("\t") -NoNewLine -ForegroundColor Cyan
	Write-Host "date" -NoNewLine
	Write-Host ("\t") -NoNewLine -ForegroundColor Cyan
	Write-Host "object" -NoNewLine
	Write-Host ("\t") -NoNewLine -ForegroundColor Cyan
	Write-Host "message)"
	Write-Host "--------------------------------------------------------------------------------------------"
	Write-Host ""
	exit
}

# SHOW EXAMPLES
if ($PSBoundParameters.ContainsKey('Examples') -and $FromConsole) {
	$examples_str = `
	"=================================================================================================" `
	+ "`nBLUE IRIS LOGFILE WATCHDOG UTILITY - USAGE EXAMPLES" `
	+ "`n--> to try these commands, copy/paste the colored text to the console prompt" `
	+ "`n--> to add simulated error entries to the Blue Iris logfile for testing this script, see TIP" `
	+ "`n-------------------------------------------------------------------------------------------------" `
	+ "`n  1. Show count of new entries (rows) added to the BI logfile since the script was last executed;" `
	+ "`n     (and, if defined, apply the filters in `$my_default_filters in the user-settings-file)" `
	+ "`n     $ScriptName"`
	+ "`n  2. Clear script's batch mode management files (use if script is not running as expected)" `
	+ "`n     $ScriptName -Reset"`
	+ "`n  3. Check if BI logfile contains new 'warning' OR 'error' entries (see TIP to simulate)" `
	+ "`n     $ScriptName -FilterLevels '1|2' -NoPush -NoBiLog"`
	+ "`n  4. Like #3, but send notification, and (if configured) post a summary to the BI logfile" `
	+ "`n     $ScriptName -FilterLevels '1|2'" `
	+ "`n  5. Send notification if new 'message' column entries have 'fail' OR 'corrupt' substrings" `
	+ "`n     $ScriptName -FilterMsgs 'fail|corrupt'" `
	+ "`n  6. Send notification if new error entries involve storage folders 'Alerts' OR 'Aux 1'" `
	+ "`n     $ScriptName -FilterLevels 2 -FilterObjects 'Alerts|Aux 1'" `
	+ "`n  7. Send notification if new error entries are for overallocated drives" `
	+ "`n     $ScriptName -FilterLevels 2 -FilterMsgs 'free','-'" `
	+ "`n     $ScriptName -FilterLevels 2 -FilterMsgs 'free%2C -' ... use %2C for literal comma" `
	+ "`n  8. Check if logfile contains new entries per the specified batch file " `
	+ "`n     $ScriptName -BatchFile 'd:\my_filters.txt' -NoPush -NoBiLog" `
	+ "`n  9. Like #8, but send notification(s), and (if configured) post summary to the BI logfile" `
	+ "`n     $ScriptName -BatchFile 'd:\my_filters.txt" `
	+ "`n 10. Use a mock logfile instead the active Blue Iris logfile (for testing)" `
	+ "`n     $ScriptName -FilterMsgs 'err|fail' -TestFile 'D:\test.txt' -NoBiLog -NoPush" `
	+ "`n-------------------------------------------------------------------------------------------------"

	# For inline hashtables, use unquoted keys and a semicolon separator
	# https://learn.microsoft.com/en-us/powershell/scripting/learn/deep-dives/everything-about-hashtable?view=powershell-7.4
	$InputObject = @{
		Colors = @{ "$ScriptName" = 'Cyan' };
		SimpleMatch = $true;
		PauseOutput = $true
	}
	# We use an object in the following pipe because 
	# REF: https://stackoverflow.com/a/53780647/13499984
	$examples_str | Format-Color $InputObject
	
	Write-Host "TIP: To add simulated error entries to the Blue Iris logfile for testing,"
	Write-Host "use URLs like following Blue Iris HTTP interface command in your internet browser:"
	Write-Host ("http://" + $bi_webserver + "/admin?log=demo%20message&level=2") -ForegroundColor Cyan
	Write-Host ""
	
	exit
}

# ------------------------------------------------------
# MANAGE SCRIPT'S FOLDERS
if (1 -eq 1) {  # Use to collapse this code section while editing

	# 1. Home folder
	if (!(Test-Path -LiteralPath $app_path)) {
		New-Item -ItemType Directory -Force -Path $app_path | Out-Null
	}
		
	# 2. Create the Script's batch mode management subfolder, if necessary
	$scr_working_folder = "_script_files"
	$scr_working_path = "$app_path\$scr_working_folder"
	if (!(Test-Path -LiteralPath $scr_working_path)) {
		New-Item -ItemType Directory -Force -Path $scr_working_path | Out-Null
	}
	# ... and add/refresh a README to the folder
	$readme_path = "$scr_working_path\_README.txt"
	if (!(Test-Path -LiteralPath $readme_path)) {
		New-Item -Path "$readme_path" | Out-Null
		$content = "Script Management Folder" +
			"`n------------------------" +
			"`n`nDESCRIPTION: This folder holds script-managed files used for:" +
			"`n1. tracking the number of lines in the Blue Iris active logfile between executions, and saving a copy of those lines" + 
			"`n2. managing recursive script executions via the -BatchFile argument" + 
			"`n`nNote: This folder and this README.txt file are programmatically created."

		# Ensure UTF-8 w/o BOM in PS5.1 & PS7 using version detection
		if ($PSVersionTable.PSVersion.Major -lt 7) {
			$utf8NoBom = New-Object System.Text.UTF8Encoding $false
			[System.IO.File]::WriteAllText($readme_path, $content, $utf8NoBom)
		} else {
			$content | Out-File -FilePath "$readme_path" -Encoding utf8NoBOM
		}
	}
}

# FILE PATHS USED BY THIS SCRIPT
# File used to save line count of the active logfile
$logfile_rows_count_path = $scr_working_path + "\active_logfile_rows_count.txt"
# File used to show the newest rows in the active logfile (-DevMode only)
$logfile_rows_newest_path = $scr_working_path + "\active_logfile_rows_newest.txt"
#
# 'Batch Mode' paths/files (used with -BatchFile argument)
# ... default path
$batchfile_arg_path = ""
# ... track which line is being executed from a -BatchFile argument
$batchfile_arg_actionable_lines = $scr_working_path + "\batchfile_arg_actionable_lines.txt"
# ... track which line is being executed from a -BatchFile argument
$batchfile_arg_status_path = $scr_working_path + "\batchfile_arg_status.txt"
# ... notify script that no new data found in the logfile from a -BatchFile argument
$batchfile_arg_no_new_entries_path = $scr_working_path + "\batchfile_arg_no_new_logfile_entries.txt"
# ... notify script to reuse import logfile lines
$batchfile_arg_newest_rows_path = $scr_working_path + "\batchfile_arg_newest_rows.txt"
# ... notify script to stop executing recursive script calls via the -BatchFile argument
$batchfile_arg_done_path = $scr_working_path + "\batchfile_arg_done.txt"
# ... collector for matched rows from executing recursive script calls via the -BatchFile argument
$batchfile_arg_matches_path = $scr_working_path + "\batchfile_arg_matches.txt"
# ... notify script to break out of loop executing recursive script calls via the -BatchFile argument
# $batchfile_arg_break_path = $scr_working_path + "\batchfile_arg_break.txt"

#----------------------------------------------
# HANDLE COMMANDLINE ARGS
# ... NOTE: must follow code section 'MANAGE SCRIPT'S FOLDERS'
if (1 -eq 1) {  # Use to collapse this code section while editing

	if ($PSBoundParameters.ContainsKey('BatchFile')) {
		if ($PSBoundParameters.ContainsKey('FilterMsgs') -or $PSBoundParameters.ContainsKey('FilterLevels') -or $PSBoundParameters.ContainsKey('FilterObjects')) {
			$err_msg = "ERR/argument -BatchFile should not be used w/ arguments -FilterMsgs, -FilterLevels -or -FilterObjects"
			return $err_msg						
		}
		if (!$BatchFile.EndsWith('.txt')) {
			$err_msg = "ERR/expecting a filename like 'my_filters.txt' when using argument -BatchFile"
			return $err_msg						
		}
	}
	
	$filter_levels_arg_ary = @()
	$filter_levels_arg_ary_cnt = 0
	if ($PSBoundParameters.ContainsKey('FilterLevels')) {
		if ($FilterLevels.count -gt 0) {
			$filter_levels_arg_ary = $FilterLevels
			$filter_levels_arg_ary_cnt = $filter_levels_arg_ary.Count
		}
	}

	$filter_objects_arg_ary = @()
	$filter_objects_arg_ary_cnt = 0
	if ($PSBoundParameters.ContainsKey('FilterObjects')) {
		if ($FilterObjects.count -gt 0) {
			$filter_objects_arg_ary = $FilterObjects
			$filter_objects_arg_ary_cnt = $filter_objects_arg_ary.Count
		}
	}

	$filter_msgs_arg_ary = @()
	$filter_msgs_arg_ary_cnt = 0
	if ($PSBoundParameters.ContainsKey('FilterMsgs')) {
		if ($FilterMsgs.count -gt 0) {
			$filter_msgs_arg_ary = @($FilterMsgs)
			$filter_msgs_arg_ary_cnt = $filter_msgs_arg_ary.Count
		}
	}
	
	$using_filter_args_flg = ($filter_levels_arg_ary_cnt -gt 0 -or $filter_objects_arg_ary_cnt -gt 0 -or $filter_msgs_arg_ary_cnt -gt 0)
	$using_batchfile_arg_flg = $PSBoundParameters.ContainsKey('BatchFile')
	if ($using_filter_args_flg -or $using_batchfile_arg_flg) {
		if ($FromConsole) {
			if ($using_my_default_filters_flg -and $line_num -eq 1) {
				#Write-Host ("Reminder: User-setting `$my_default_filters is ignored when using arguments -Filter* or -Batchfile") -ForegroundColor Green 
			}
		}
		$using_my_default_filters_flg = $false		
	} else {
		if ($FromConsole) {
			if ($using_my_default_filters_flg) {
				# Write-Host ("Using `$my_default_filters from the user-settings-file") -ForegroundColor Green 
			}
		}		
	}
	
	# For debugging
	# Write-Host "filter_levels_arg_ary_cnt = $filter_levels_arg_ary_cnt"
	# Write-Host "filter_levels_arg_ary = $filter_levels_arg_ary"
	# Write-Host "filter_objects_arg_ary_cnt = $filter_objects_arg_ary_cnt"
	# Write-Host "filter_objects_arg_ary = $filter_objects_arg_ary"
	# Write-Host "filter_msgs_arg_ary_cnt = $filter_msgs_arg_ary_cnt"
	# Write-Host ("filter_msgs_arg_ary = " + ($filter_msgs_arg_ary | ConvertTo-Json))
	# Write-Host "using_filter_args_flg = $using_filter_args_flg"
	# Write-Host "using_batchfile_arg_flg = $using_batchfile_arg_flg"
	# Write-Host "using_my_default_filters_flg = $using_my_default_filters_flg"
	
	if ($Renew) {
		# Remove orphaned recursive call management files
		if (Test-Path -Path "$scr_working_path\*" -Include "active_logfile_rows*.txt") {
			Remove-Item -Path "$scr_working_path/active_logfile_rows*.txt"
			if ($FromConsole) {
				Write-Host ("Deleted active logfile tracking files in '$app_path\_script_files'") -ForegroundColor Green
			}
		}
	}

	if ($Reset) {
		# Remove orphaned recursive call management files
		if (Test-Path -Path "$scr_working_path\*" -Include "batchfile_arg_*.txt") {
			KillRecursiveMgtFiles -RootPath "$scr_working_path"
			if ($FromConsole) {
				Write-Host ("Deleted -BatchFile mode recursive call management files in '$app_path\_script_files'") -ForegroundColor Green
			}
		}
		if ($using_filter_args_flg -or $using_batchfile_arg_flg) {
			$filter_levels_arg_ary = @()
			$filter_objects_arg_ary = @()
			$filter_msgs_arg_ary = @()
			$using_filter_args_flg = $false
			$BatchFile = ""
			$using_batchfile_arg_flg = $false
			if ($FromConsole) {
				Write-Warning ("arguments -Filter*, -BatchFile ignored when using argument -Reset")
			}
		}
	}
	
	# Make string of params (used for logging to the script log & the Blue Iris logfile)
	$params_ary = @()
	foreach ($key in $PSBoundParameters.Keys){
		$value = (get-variable $key).Value 
		if ($value -eq $true) {
			$params_ary += "-$key"
		} else {
			$params_ary += "-$key $value"
		}
	}
	$params_str = $params_ary -Join " "
	# Write-Host "params_str = $params_str"
	# Write-Host ("PSBoundParameters cnt = " + $PSBoundParameters.Keys.Count)
	# Write-Host ("PSBoundParameters = " + $PSBoundParameters.Keys)
	# Write-Host ("params_ary = " + ($params_str | ConvertTo-Json))

	if ($PSBoundParameters.ContainsKey('PushService')) {
		if (!(@("pushover", "telegram") -contains $PushService.ToLower())) {  # -contains operator is case-insensitive by default
			$err_msg = "ERR/argument -PushService must be 'Pushover' or 'Telegram'"
			return $err_msg
		} else {
			$push_service = $PushService.ToLower()
		}
	}
	# Write-Host "push_service = '$push_service'"

	if ($PSBoundParameters.ContainsKey('Title')) {
		$notification_title = $Title
	} else {
		$notification_title = "via $ScriptName"
	}

	if ($PSBoundParameters.ContainsKey('Msg')) {
		$notification_msg = $Msg
	} else {
		$notification_msg = ""
	}
	
	if ($PSBoundParameters.ContainsKey('Priority')) {
		if (($Priority -lt -2) -or ($Priority -gt 2)) {
			$err_msg = "ERR/expecting -Priority between -2 and 2"
			return $err_msg
		}
		$notification_priority = $Priority
	} else {
		$notification_priority = $po_priority
	}
	# Write-Host "notification_priority = $notification_priority"
		
	if ($PSBoundParameters.ContainsKey('Sound')) {
		$notification_sound = $Sound
	} else {
		$notification_sound = $po_sound
	}
	
	if ($PSBoundParameters.ContainsKey('Devices')) {
		$notification_devices = $Devices
	} else {
		$notification_devices = $po_device
	}
	
	# $test_file_mode_flg = $false
	if ($PSBoundParameters.ContainsKey('TestFile')) {
		if (!(Test-Path $TestFile)) {
			# Try to find it the $app_folder
			if (Test-Path "$app_path\$TestFile") {
				$TestFile = $app_path + "/" + $TestFile
			} else {
				$err_msg = "ERR/non-existent file in argument -TestFile ('$TestFile')"
				return $err_msg
			}
		}
		if ($PSBoundParameters.ContainsKey('BatchFile')) {
			# $test_file_mode_flg = $true
			# Remove orphaned recursive call management files
			<# KillRecursiveMgtFiles -RootPath "$scr_working_path"
			if ($FromConsole) {
				Write-Host ("Argument -Reset is automatically applied when -TestFile is used with -BatchFile") -ForegroundColor Green
				Write-Host ("Deleted -BatchFile mode recursive call management files in '$app_path\_temp'") -ForegroundColor Green
			} #>
			#exit
		}

		# Clear batch mode management files
		if (Test-Path -Path "$RootPath\*" -Include "batchfile_arg_*.txt") {
			KillRecursiveMgtFiles -RootPath "$scr_working_path"
			if ($FromConsole) {
				Write-Host ("Argument -Reset is automatically applied when using -TestFile") -ForegroundColor Green
				Write-Host ("Deleted -BatchFile mode recursive call management files in '$app_path\_temp'") -ForegroundColor Green
			}
		}
		
	}
}

if ($FromConsole -and $DevMode) {
	if ($using_filter_args_flg) {
		Write-Host "`n"
	}
	Write-Host ("(dev mode) entered script...") -ForegroundColor Cyan
}

# Remove previously used recursive call management files
if ($using_filter_args_flg -and !(Test-Path -LiteralPath $batchfile_arg_status_path)) {
	KillRecursiveMgtFiles -RootPath "$scr_working_path"
} 

# PROCESS THE -BatchFile ARGUMENT FILE -OR- USER-SETTINGS-FILE VARIABLE $my_default_filters  
if ($using_batchfile_arg_flg -or $using_my_default_filters_flg) {

	# Remove previously used recursive call management files
	KillRecursiveMgtFiles -RootPath "$scr_working_path"

	# Initialize LIST collector for actionable lines in the -AttachFile arg filespec
	$actionable_lines = New-Object System.Collections.Generic.List[System.Object]

	# Set of substrings, of which one, must exist in an actionable_line
	$accepted_batch_args = @('-FilterLevels','-FilterObjects','-FilterMsgs','-PushService','-Title','-Priority','-Sound','-Devices')

	if ($using_my_default_filters_flg) {
		
		$batchfile_arg_path = 'user-setting-my_default_filters'
		
		# Use the filter set from the user-settings-file
		$contents = $my_default_filters
		
		if ($FromConsole) {
			Write-Host ("Using `$my_default_filters from the user-settings-file") -ForegroundColor Green
		}

	} else {
		# Import the user batch file
		# ... make the filespec
		if ($BatchFile -match ":" -or $BatchFile -match "\\\\") {
			# ... the requested file is in local filesystem, or network share
			$batchfile_arg_path = $BatchFile
		} else {
			# ... the requested file is in this scripts working folder
			$batchfile_arg_path = "$app_path\$BatchFile"
		}
		
		# For testing
		# $batchfile_arg_path += "1" 

		if (Test-Path -LiteralPath $batchfile_arg_path) {
			$contents = @(Get-Content -LiteralPath $batchfile_arg_path | ForEach-Object { $_.ToString() })
		} else {
			$err_msg = "ERR/non-existent file in argument -BatchFile ('$batchfile_arg_path')"
			return $err_msg				
		}
	}
	
	# Expecting multi-line file with contents like...
	#   1. # comment
	#	2. FilterMsgs="admin|DB"
	#   3. # comment
	#	4. FilterObjects="admin";FilterMsgs="login|logout"
	#   5. <blank>
	#   6. # comment
	#	7. FilterLevels="1|2";FilterObjects="admin";FilterMsgs="login|logout"

	# For debugging
	# Write-Host ("contents type = " + $contents.GetType().Name)
	# Write-Host ("contents = " + ($contents | ConvertTo-Json))
	
	# Exclude blank and commented lines
	# ... commented lines may start with '#' or '/' or '//'
	foreach ($line in $contents) {
		if ($line) { 
			#Write-Host "line = $line"
			if (!($line.StartsWith('#') -or $line.StartsWith('/'))) {
				foreach ($expected_arg in $accepted_batch_args) {
					if ($line -like "*$expected_arg*") {
						$actionable_lines.Add($line)
						break
					}
				}
			}
		}
	}
	$lines_cnt = $actionable_lines.Count

	# For debugging
	# Write-Host ("actionable_lines (w/o comments) type = " + $actionable_lines.GetType().Name)
	# Write-Host "actionable_lines (w/o comments) cnt = $lines_cnt"
	# Write-Host ("actionable_lines (w/o comments) = " + ($actionable_lines | ConvertTo-Json))
	
	if ($FromConsole -and $DevMode) {
		Write-Host ("(dev mode) -BatchFile arg file '$BatchFile' has " + ($actionable_lines.Count) + " actionable line(s)") -ForegroundColor Cyan
		Write-Host ($actionable_lines | ConvertTo-Json) -ForegroundColor Cyan
	}
	
	# Save the actionable lines to TXT file (useful for debugging, etc.)
	$content = $actionable_lines -Join "`n"
	Set-Content -LiteralPath $batchfile_arg_actionable_lines -Value $content

	# ITERATE BATCH FILE LINES (MULTIPLE FILTERS)
	if ($actionable_lines.Count -gt 0) {

		# SUCCESS - ARGUMENT -BatchFile VALIDATED

		if ($FromConsole -and $DevMode) {
			# Write-Host ("(dev mode) deleted old -BatchFile argument management files") -ForegroundColor Cyan
			Write-Host ("(dev mode) deleted old recursive mode management files in subfolder '_temp' (used w/ -BatchFile argument)") -ForegroundColor Cyan
		}
		
		# RECURSIVELY EXECUTE EACH ACTIONABLE LINE
		# Expecting LIST of $actionable_lines to be like...
		# @(
		#	"-FilterMsgs=\"login|logout\""
		#	"-FilterObjects=\"admin\";FilterMsgs=\"login|logout\""
		#	"-FilterLevels=\"10|2\";-FilterObjects=\"admin\";-FilterMsgs=\"login|logout\""
		# )

		$line_num = 1
		foreach ($line in $actionable_lines) {
			
			# $line is like...
			# ... "-FilterObjects='App','DB';-FilterMsgs='login|logout','err'"
			# ... "-FilterObjects='App','DB';-FilterMsgs='login|logout','err';-Priority=1"
			
			# Break out of loop if there we no new lines in the Blue Iris logfile
			if ($line_num -gt 1) {
				$content = @(Get-Content -LiteralPath $logfile_rows_newest_path | ForEach-Object { $_.ToString() }) 
				if (!$content) {
					break
				}
			}

			if ($FromConsole -and $DevMode) {
				Write-Host ""
				Write-Host ("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~") -ForegroundColor Cyan
				Write-Host ("(dev mode) processing -BatchFile actionable line $line_num/$lines_cnt") -ForegroundColor Cyan
				Write-Host ("(dev mode) --> line: " + $line) -ForegroundColor Cyan
			}
			
			# Get the arguments requested
			$arr = $line.Split(';')
			#  ... just in case, fix inadvertent spaces around ';' separators
			$line_bits = $arr | ForEach-Object { $_.Trim() } 
			
			# if $line is like 
			# -FilterLevels="10";-FilterObjects="admin|local_console";-FilterMsgs="login|logout"
			# then $line_bits is like...
			# [
			#	"-FilterLevels=\"10\"",
			#	"-FilterObjects=\"admin|local_console\"",
			#	"-FilterMsgs=\"login|logout\""
			#	"-PushServoce=\"telegram\""
			#	"-Priority=1"
			#	"-Sound=\"sound\""
			#	"-Devices=\"device1,device2\""
			# ]
			
			# For debugging
			# Write-Host ("line_bits = " + ($line_bits | ConvertTo-Json))
			
			# Initialize accumulators - to be used with call operator below
			# WHY? this is the only construct that successfully passes the arguments via the call operator
			$args_ht = [Ordered]@{}  
			
			foreach ($arg_pair in $line_bits) {
				
				# $arg_pair is like...
				# ... "-FilterObjects=\"App\",\"DB\""
				# $arg_pair_parts are like...  
				# ... @( "-FilterObjects", "\"App\",\"DB\"" }
				# KEY POINT: note the inner double-quotes are escaped by PowerShell
				
				$arg_pair_parts = $arg_pair.Split('=')
				
				# For debugging
				# Write-Host ("arg_pair_parts = " + ($arg_pair_parts | ConvertTo-Json))
				
				if ($arg_pair_parts.count -eq 2) {  # Skip malformed lines
					
					$arg_str = $arg_pair_parts[0]  # like '-FilterObjects'
					
					# Make sure the requested argument is supported
					if (!($accepted_batch_args -contains $arg_str)) {
						
						if ($FromConsole) {
							Write-Warning ("ignoring unsupported entry ('$arg_str') in argument -BatchFile line'`n--> '$line'")
						}

					} else {
	
						# PREP REQUESTED ARGUMENT SUBSTRING FOR RECURSIVE CALLS TO THIS SCRIPT

						# Add the argument to the hashtable to be using with the CALL OPERATOR LINE below
						$key = $arg_str

						# Get the variable object associated with the $key
						$key_var_name = $key.TrimStart('-') 
						$key_var_obj = Get-Variable -Name $key_var_name

						# Get the type of the variable
						if ($null -eq $key_var_obj.Value) {
							$key_var_value_type = "null"
						} else {
							$key_var_value_type = $key_var_obj.Value.GetType().Name
						}						
						
						# For debugging
						#$key_var_obj
						# Write-Host ("key [$" + $key_var_name + "] obj =`n" + ($key_var_obj | ConvertTo-Json))
						# Write-Host ("key [$" + $key_var_name + "] type = $key_var_value_type")
						
						if ($key -match '-Priority') {
						#if ($key_var_value_type -match 'Int') {  # Generalized
							
							$val = [Int]($arg_pair_parts[1].Trim('"').Trim("'"))
							$args_ht[$key] = $val

							# For debugging
							# Write-Host ("args_ht[" + $key + "] type = " + $val.GetType().Name)

						} elseif (@('-Title','-PushService','-Priority','-Devices','-Sound') -contains $key) {
							
							$val = $arg_pair_parts[1].Trim('"').Trim("'")
							$args_ht[$key] = $val

						} else {
							
							#$val = $arg_pair_parts[1].Trim('"').Trim("'")
							$val = $arg_pair_parts[1].Split(',')
							$args_ht[$key] = @($val)  # MUST force single strings to an array for splatting
						}

						# For debugging
						# Write-Host ("args_ht[" + $key + "] = " + $val)

					}
				}
			}
			
			# For debugging
			# Write-Host ("args_ht keys cnt = " + $args_ht.Keys.Count)
			# Write-Host ("args_ht = " + ($args_ht | ConvertTo-Json))
			
			# For testing
			#$args_ht = @{}
			
			if ($args_ht.Keys.Count -gt 0) {
				
				# Add original call's optional arguments to the hashtable
				#$args_ht["-Reset"] = $Reset
				$args_ht["-ShowMatches"] = $ShowMatches
				$args_ht["-NoPush"] = $NoPush
				$args_ht["-NoBiLog"] = $NoBiLog
				$args_ht["-Quiet"] = $Quiet
				$args_ht["-DevMode"] = $DevMode
				if (!$args_ht["-Title"]) {
					$args_ht["-Title"] = $notification_title
				}					
				if (!$args_ht["-PushService"]) {
					$args_ht["-PushService"] = $push_service
				}					
				if (!$args_ht["-Priority"]) {
					$args_ht["-Priority"] = $notification_priority
				}					
				if (!$args_ht["-Devices"]) {
					$args_ht["-Devices"] = $notification_devices
				}					
				if (!$args_ht["-Sound"]) {
					$args_ht["-Sound"] = $notification_sound
				}
				if ($TestFile) {		
					if (!$args_ht["-TestFile"]) {
						$args_ht["-TestFile"] = $TestFile
					}
				}					
			
				# For debugging
				# Write-Host ("args_ht keys cnt = " + $args_ht.Keys.Count)
				# Write-Host ("args_ht = " + ($args_ht | ConvertTo-Json))

				# Create temp file to notify script to reuse the newest logfile lines
				# WHY? to avoid repeated queries of the active logfile, 
				#      which can continued to add lines during the script's execution
				$content = $batchfile_arg_path
				$content += "`n" + $lines_cnt.ToString()
				$content += "`n" + $line_num.ToString()
				Set-Content -LiteralPath $batchfile_arg_status_path -Value $content

				if ($line_num -eq 1) {
					$verb = "initialized"
				} else{
					$verb = "updated"
				}

				# For debugging
				# Write-Host "'$batchfile_arg_status_path' new content = `n$content"
				# Write-Host "$verb '$batchfile_arg_status_path'"

				if ($FromConsole -and $DevMode) {
					Write-Host ("(dev mode) --> $verb recursive mode tracking file: $batchfile_arg_status_path") -ForegroundColor Cyan
					#Write-Host ("(dev mode) --> hashtable for pending recursive call:`n" + ($args_ht | ConvertTo-Json)) -ForegroundColor Cyan
					Write-Host ("(dev mode) --> hashtable for pending recursive call:") -ForegroundColor Cyan
					#$Host.UI.RawUI.ForegroundColor = "Cyan"
					$args_ht
					#$Host.UI.RawUI.ForegroundColor = $OriginalForegoumdColor					
				}
				
				# ==============================
				# RECURSIVELY CALL THIS SCRIPT
				# ==============================
				try {
					
					# For debugging
					# Write-Host "`n------------------------------------"
					# Write-Host ("Starting Filter Action #" + $line_num + "...")
					# Write-Host "recursive execution arguments (using 'splatting'):"
					# $args_ht

					# CALL OPERATOR LINE
					# KEY POINT: the only way I got this to work was using a hashtable and "splatting" (@args_ht)
					# REF: https://stackoverflow.com/a/52587519/13499984
					# None of these approaches worked:
					#	& $ScriptName $args_str  # FAIL
					#	Invoke-Expression ".$ScriptName $args_str"  # FAIL
					#	& $ScriptName $args_ary  # FAIL
					#	& $ScriptName @args_ary  # FAIL
					& $ScriptName @args_ht

					# For debugging
					Write-Host ("Success! - done W/ Filter Action #" + $line_num)
					# Write-Host ("Arguments hashtable used:") -ForegroundColor Green
					# $args_ht

				}
				catch {
					if ($FromConsole) {
						Write-Warning ("unable to recursively call this script using the arguments set in -BatchFile actionable line #" +  $line_num)
						Write-Warning ("line #" + $line_num + ": " + $line)
						Write-Warning ("arguments hashtable attempted:")
						$args_ht
					} else {
						$err_msg = ("unable to recursively call this script using the arguments set in -BatchFile actionable line #" + $line_num)
						return $err_msg
					}
				}
				
				$line_num++
				
			} else {
				$err_msg = "Argument -BatchFile contains no accountable lines"
				return $err_msg
			}
		}
		
		# Finished processing -BatchFile
		$rtn_ary = @()
		if (Test-Path -LiteralPath $batchfile_arg_matches_path) {
			$content = $matched_rows -Join "`n"
			$arr = @(Get-Content -LiteralPath $batchfile_arg_matches_path | ForEach-Object { $_.ToString() })
			$rtn_ary = $arr | Get-Unique  # Remove duplicate entries
		}
		if ($FromConsole) {
			if ($DevMode) {
				Write-Host ("====================") -ForegroundColor Cyan
				Write-Host ("(dev mode) batch mode rtn_ary =`n" + ($rtn_ary | ConvertTo-Json)) -ForegroundColor Cyan
			}
		} else {
			return $rtn_ary
		}
		#exit
	}

}

# Initialize vars used when script is being recursively called (via the -BatchFile argument)
$recursive_mode_flg = $false  # assumed $false until verified by status file
$batchfile_arg_active_line_num = 0
$batchfile_arg_line_num_is_last = (Test-Path -LiteralPath $batchfile_arg_done_path)
$batchfile_arg_no_new_entries_flg = (Test-Path -LiteralPath $batchfile_arg_no_new_entries_path)

# ABORT SCRIPT if no new entries found in the logfile and script is being recursively called via the -BatchFile argument
if ($batchfile_arg_no_new_entries_flg) {
	
	# For debugging
	# Write-Host "no new entries found in the logfile and script is being recursively called via the -BatchFile argument"
	
	$rtn_ary = @()
	return $rtn_ary
}

# GET RECURSIVE MODE STATUS (i.e., user requested the -BatchFile argument)
if (1 -eq 1) {  # Use to collapse this code section while editing
	if ((Test-Path -LiteralPath $batchfile_arg_status_path) -and !$batchfile_arg_line_num_is_last) {
		
		# NOTE: the @() wrapper in the next line is unnecessary in this case b/c 
		#       we know for certain that the file's contents are never a single line
		#$status_ary = @(Get-Content -LiteralPath $batchfile_arg_status_path)
		$status_ary = Get-Content -LiteralPath $batchfile_arg_status_path
		
		# Expecting...
		# status_ary = @(
		#   [0] path to file to in -BatchFile argument (or 'user-setting-my_default_filters')
		#   [1] total number of actionable lines
		#   [2] line number of current actionable line
		#@ )

		# For debugging
		# Write-Host ("status file type = " + $status_ary.GetType().Name)
		# Write-Host ("status_ary = " + ($status_ary -Join ", "))
		if ($PSVersion -gt 5) {
			# We avoid the following line in PowerShell 5 b/c it will also 
			# show the Get-Content cmdlet's PSObject metadata
			# Write-Host ("status_ary = `n" + ($status_ary | ConvertTo-Json))  
		} else {
			# Write-Host ("status_ary = " + ($status_ary -Join ", "))
		}
		# Write-Host ("Retrieved recursive mode status data:`n" + ($status_ary -Join "`n"))

		if ($status_ary.Count -ge 1) {
			
			# 1st line is the total number of lines in the -BatchFile argument
			$filterfile_path = $status_ary[0]
			
			# 2nd line is the total number of lines in the -BatchFile argument
			$batchfile_arg_lines_cnt = [Int]$status_ary[1]

			# last line is the active line# in the -BatchFile argument
			$batchfile_arg_active_line_num = [Int]$status_ary[2]

			if ($batchfile_arg_active_line_num -eq 1) {
				if ($FromConsole -and !$Quiet) {
					if ($filterfile_path -notmatch '.txt') {
						Write-Host "`nEXECUTING MULTIPLE FILTER ACTIONS VIA `$my_default_filters ..."
						Write-Host "============================================================="
					} else {
						Write-Host "`nEXECUTING MULTIPLE FILTER ACTIONS VIA -BatchFile ..."
						Write-Host "===================================================="
					}
				}
			}

			if ($batchfile_arg_active_line_num -gt 1) {
				$recursive_mode_flg = $true
			}
				
			# Create FLAG file if we're processing the final entry in the -BatchFile argument
			if ($batchfile_arg_active_line_num -eq $batchfile_arg_lines_cnt) {
				Set-Content -LiteralPath $batchfile_arg_done_path -Value "completed the last entry in 'filterfile_path'"
			}
		}
	}

	# For debugging
	# Write-Host ("-BatchFile arg is reusing data: $recursive_mode_flg")
	# Write-Host "batchfile_arg_active_line_num = $batchfile_arg_active_line_num"

	# # For testing
	# $batchfile_arg_active_line_num = 2
	# if ($batchfile_arg_active_line_num -gt 1) {
		# exit
	# }

	# Now that we have the recursive mode status...
	if ($FromConsole) {
		# ... show this reminder
		if ($batchfile_arg_active_line_num -eq 0) {
			$test1 = ($my_default_filters.Count -gt 0)
			$test2 = ($filter_levels_arg_ary_cnt -gt 0 -or $filter_objects_arg_ary_cnt -gt 0 -or $filter_msgs_arg_ary_cnt -gt 0)
			$test3 = $PSBoundParameters.ContainsKey('BatchFile')
			if ($test1 -and $test2 -and !$test3) {
				Write-Host ("Reminder: User-setting `$my_default_filters is ignored when using arguments -Filter* or -Batchfile`n") -ForegroundColor Green 
			}
		}
	}
}

# Start timer
$timer = [Diagnostics.Stopwatch]::StartNew()

# REUSE LOGFILE DATA WHEN MAKING 2ND, 3RD... RECURSIVE CALLS TO THIS SCRIPT
if ($recursive_mode_flg) {
	
	# For debugging
	# Write-Host ">> Using saved logfile data (previously retrieved new lines) <<"
	
	# For testing
	#$batchfile_arg_newest_rows_path += "1"
	
	# USE THE PREVIOUSLY SAVED DATA

	if (Test-Path -LiteralPath $batchfile_arg_newest_rows_path) {
		
		# Import the logfile's previously saved newest rows to an array
		
		# Originally we used the next line, but it did not work as expected in PowerShell ISE (V5.1)
		#$newest_rows = @(Get-Content -LiteralPath $batchfile_arg_newest_rows_path)
		
		# Per ChatGPT, in PowerShell ISE, the Get-Content cmdlet returns file information as well as the desired content. 
		# To avoid having Get-Content return objects rather than just the content lines, we use ForEach-Object { $_.ToString() }
		$newest_rows = @(Get-Content -LiteralPath $batchfile_arg_newest_rows_path | ForEach-Object { $_.ToString() })

		if ($null -eq $newest_rows) {
			$newest_rows_cnt = 0
		} else {
			$newest_rows_cnt = $newest_rows.Count
		}

		if ($FromConsole -and $DevMode) {
			Write-Host ("(dev mode) using previously retrieved & saved logfile data ($newest_rows_cnt rows)") -ForegroundColor Cyan
		}

	} else {
		Set-Content -LiteralPath $batchfile_arg_done_path -Value "completed the last entry in 'filterfile_path'"
		$err_msg = "ERR/unable to find the saved logfile data for argument -BatchFile ('\" + $scr_working_folder + "\" + (Split-Path $batchfile_arg_newest_rows_path -Leaf) + "')"
		return $err_msg	
	}
}

# GET LOGFILE'S NEWEST ENTRIES (FOR STANDARD EXECUTION -OR- FOR THE 1ST TIME WHEN USING RECURSIVE MODE)
if (!$recursive_mode_flg) {

	if ($PSBoundParameters.ContainsKey('TestFile')) {
		
		# For debugging
		# Write-Host ">> Retrieving the newest entries (rows) from -TestFile <<"
		
		$bi_logfile_filespec = $TestFile
		$bi_logfile_name = Split-Path $TestFile -Leaf

		# TEST FOR MISSING OR MALFORMED HEADER
		# ... automatically fix the file if necessary
		$header_row = Get-Content -LiteralPath $bi_logfile_filespec -TotalCount 1
		$columns = $header_row -split "`t"  # Split the line by tabs
		# test1 is true if the file's 1st row is a good header
		# ... if output of compare-object is null, this means the two arrays are identical.
		$test1 = ($null -eq (Compare-Object $columns @("level","time","object","message")))
		# test2 helps us detect if the file has a malformed header
		$test2 = ($header_row -match "level" -and $header_row -match "time" -and $header_row -match "object" -and $header_row -match "message")
		#Write-Host "test1 = $test1, test2 = $test2"
		if (!$test1) {
			$header = "level`t" + "time`t" + "object`t" + "message"
			if (!$test2) {  # first row is NOT a malformed header <== insert a header
				$fileContent = Get-Content -Path "$bi_logfile_filespec"
				$newContent = @($header) + $fileContent  # merge arrays
				$newContent | Set-Content -Path "$bi_logfile_filespec"

				if ($FromConsole) {
					Write-Host ("HEADS-UP: inserted missing header to -TestFile ($bi_logfile_name)") -ForegroundColor Green
					Write-Host ("(new header: '" + $header + "')") -ForegroundColor Green
				}
			} else {  # first row is malformed (look-alike) header <== replace it
				$fileContent = Get-Content -Path "$bi_logfile_filespec"
				$fileContent[0] = $header
				$fileContent | Set-Content -Path "$bi_logfile_filespec"

				if ($FromConsole) {
					Write-Host ("HEADS-UP: replaced malformed header in -TestFile ($bi_logfile_name)") -ForegroundColor Green
					Write-Host ("(original header: '" + $header_row + "')") -ForegroundColor Green
					Write-Host ("(expected header: '" + $header + "')") -ForegroundColor Green
				}
			}
		}
		
	} else {
		
		# For debugging
		# Write-Host ">> Retrieving the newest entries (rowS) from the active logfile <<"

		# GET THE ACTIVE LOGFILE
		# NOTES: 1) The Blue Iris logfile is tab-delimited
		#        2) The Blue Iris logfile columns are: "level", "time", "object", "message"
		#        3) The Blue Iris logfile did not have a header row until v5.5.5.0
		#        4) The Blue Iris logfile's 'object' column value is always 20 character long, right-padded with spaces, if needed 
		$bi_logfile_obj = Get-ChildItem -Path "$bi_logfiles_path" | Sort-Object -Property LastWriteTime -Descending | Select-Object -First 1
		$bi_logfile_filespec = $bi_logfile_obj.FullName
		$bi_logfile_name = $bi_logfile_obj.Name

		# For debugging
		# Write-Host ("bi_logfile_obj = " + ($bi_logfile_obj | ConvertTo-Json))
	
	}
	
	# For debugging
	# Write-Host "bi_logfile_filespec = '$bi_logfile_filespec'"
	
	# For testing
	#$bi_logfile_filespec += "1"

	# MAKE SURE THIS IS AN ACTUAL BLUE IRIS LOGFILE
	# ... 1. Expected file extension
	if (!$bi_logfile_filespec.EndsWith('.txt')) {
		$err_msg = "ERR/unable to validate the logfile - expecting extension '.txt' ('$bi_logfile_filespec')"
		return $err_msg	
	}
	# ... 2. Expected header
	#        KEY POINT: Blue Iris logfiles BEFORE version 5.5.5.0 did not have a header
	$logfile_is_modern = $false  # Default assumption	
	if (IsBiVersion '5.5.5.0') {
		$header_row = Get-Content -LiteralPath $bi_logfile_filespec -TotalCount 1
		# Split the line by tabs
		$columns = $header_row -split "`t"
		
		# For debugging
		# Compare-Object $columns @("level","time","object","message")

		# If output of compare-object is null, this means the two arrays are identical.
		if ($null -ne (Compare-Object $columns @("level","time","object","message"))) {
			$err_msg = "ERR/unable to validate the logfile - expecting header 'level,time,object,message' (`$bi_logfile_filespec')"
			return $err_msg	
		}
		
		$logfile_is_modern = $true
	}
	# For debugging
	# Write-Host "logfile_is_modern = $logfile_is_modern"

	if ($batchfile_arg_line_num_is_last) {	
		exit
	}

	# GET THE LOGFILE'S CURRENT LINE COUNT
	$all_rows_cnt = GetLogfileLineCount -FilePath $bi_logfile_filespec
		
	# For debugging
	# Write-Host "all_rows_cnt = $all_rows_cnt"

	# DETERMINE NUMBER OF NEW ROWS SINCE LAST CHECKED
	if ($logfile_is_modern) {
		$delta_rows = $all_rows_cnt - 1
	} else {
		$delta_rows = $all_rows_cnt
	}
	if (!$PSBoundParameters.ContainsKey('TestFile')) {
		if (Test-Path -LiteralPath $logfile_rows_count_path) {
			$header_row = Get-Content -LiteralPath $logfile_rows_count_path -TotalCount 1
			$arr = $header_row.Split('^')
			# Writes-Host ("arr = " + ($arr | ConvertTo-Json))
			if ($arr.Count -eq 2) {
				if ($arr[0] -eq $logfile_rows_count_path) {
					# Write-Host ("arr[1] = " + [Int]$arr[1])
					$delta_rows = $all_rows_cnt - [Int]$arr[1]
				}
			}
		}
	}
		
	# Update the working file used to save the current new rows count
	$new_content = "$logfile_rows_count_path" + "^" + $all_rows_cnt
	Set-Content -LiteralPath $logfile_rows_count_path -Value $new_content
		
	# For debugging
	# Write-Host "delta_rows = $delta_rows"

	if ($delta_rows -gt 0) {
		if ($FromConsole) {
			if (!$Quiet) {
				Write-Host "`nLogfile status:"
				Write-Host "---------------"
			}
			if ($PSBoundParameters.ContainsKey('TestFile')) {
				Write-Host "Detected $delta_rows entries in -TestFile ($bi_logfile_name)"
			} else {
				Write-Host "Detected $delta_rows new entries added to the logfile ($bi_logfile_name) since the last query"
			}
		}
	}

	# IMPORT THE LOGFILE'S NEWEST ROWS TO AN ARRAY
	if ($delta_rows -gt 0) {
		# Originally we used the next line, but it did not work as expected in PowerShell ISE (V5.1)
		#$newest_rows = @(Get-Content -LiteralPath $bi_logfile_filespec -Tail $delta_rows)
		
		# Per ChatGPT, in PowerShell ISE, the Get-Content cmdlet returns file information as well as the desired content. 
		# To avoid having Get-Content return objects rather than just the content lines, we use ForEach-Object { $_.ToString() }
		$newest_rows = @(Get-Content -LiteralPath $bi_logfile_filespec -Tail $delta_rows | ForEach-Object { $_.ToString() })		
	}
	if ($null -eq $newest_rows) {
		$newest_rows_cnt = 0
	} else {
		$newest_rows_cnt = $newest_rows.Count
	}
	
	if ($newest_rows) {

		# For debugging
		# Write-Host "newest_rows_cnt = $newest_rows_cnt"
		# Write-Host ("newest_rows type = " + $newest_rows.GetType().Name)
		# Write-Host ("newest_rows = " + ($newest_rows | ConvertTo-Json))
		# Write-Host ("newest_rows[0] = " + $newest_rows[0])
		# Write-Host ("newest_rows[1] = " + $newest_rows[1])
		# Write-Host ("newest_rows[" + ($newest_rows_cnt - 1) + "] = " + $newest_rows[$newest_rows_cnt - 1])
		# Write-Host ("newest_rows[0] = " + (($newest_rows[0] -Split '\t') | ConvertTo-Json))
		# Write-Host ("newest_rows[1] = " + (($newest_rows[1] -Split '\t') | ConvertTo-Json))
		# Write-Host ("newest_rows[" + ($newest_rows_cnt - 1) + "] = " + (($newest_rows[$newest_rows_cnt - 1] -Split '\t') | ConvertTo-Json))

		# Show most recent logfile rows
		if ($FromConsole -and $DevMode) {
			if (Test-Path -LiteralPath $logfile_rows_newest_path) {
				$prev_newest_rows = @(Get-Content -LiteralPath $logfile_rows_newest_path | ForEach-Object { $_.ToString() })
				$prev_newest_rows_cnt = $prev_newest_rows.count
				if ($prev_newest_rows) {
					if ($prev_newest_rows_cnt -gt 0) {
						Write-Host ("(dev mode) Previous most recent new row: `n" + $prev_newest_rows[$prev_newest_rows_cnt - 1]) -ForegroundColor Cyan
					}
				}
			}
			Write-Host ("(dev mode) Most recent new row: `n" + $newest_rows[$newest_rows_cnt - 1]) -ForegroundColor Cyan
		}

		# Save the data to a file to be reused during recursing
		if (Test-Path -LiteralPath $batchfile_arg_status_path) {
			Set-Content -LiteralPath $batchfile_arg_newest_rows_path -Value $newest_rows
		}

		# Save to working file
		Set-Content -LiteralPath $logfile_rows_newest_path -Value $newest_rows

	} else {
		# Clear working file
		Set-Content -LiteralPath $logfile_rows_newest_path -Value ""
	}
}

# APPLY FILTERS -AND- SEND NOTIFICATIONS

# ... Initialize accumulators
$matched_rows = @()
$used_levels = @() # used to highlight any matched level numbers in push notification messages
$used_terms = @()  # used to highlight any matched terms in push notification messages
$pure_or_query_ht = @{}  # hashtable for single filters that have only 'OR' logical operators

# ... Initializes
$matched_rows_cnt = 0
$found_matches_flg = $false
$pure_or_query_flg = $false
$malformed_row_num = 0

# ... Process logfile's newest rows
if ($newest_rows_cnt -gt 0) {

	# APPLY FILTERS
	if ($using_filter_args_flg) {		

		$row_num = 1
		foreach($row in $newest_rows) {
			
			# Split the row/line by tabs
			$columns = $row -Split '\t'

			# For debugging
			# Write-Host ("row #$row_num type = " + $row.GetType())
			# Write-Host "row #$row_num = $row"
			# Write-Host ("columns cnt = " + $columns.Count)
			# Write-Host ("columns = " + ($columns | ConvertTo-Json))

			# Check if the row/line has 4 columns
			if ($columns.Count -eq 4) {
				
				$level   = [Int]($columns[0].ToString().Trim())  # level is in the 1st column
				$date    = $columns[1].Trim()  # date is in the 2nd column
				$object  = $columns[2].Trim()  # object is in the 3rd column
				$message = $columns[3]         # message is in the 4th column
				
				# For debugging
				# Write-Host ""
				# Write-Host ("level = $level, type = " + $level.GetType().Name)
				# Write-Host "date = $date"
				# Write-Host "object = $object"
				# Write-Host "message = $message"
				
				# Skip this row if was created by this script
				# ... WHY? the row's 'message' column value may contain the very terms we're trying to match!
				#if ($level -eq $custom_BI_logfile_level -and $message -match $ScriptBaseName) {
				if ($message -match $ScriptBaseName) {
					continue
				}

				# CHECK IF THE ROW CONTAINS ANY OF THE SPECIFIED COLUMN VALUES...

				# ... COLUMN #1 'level'
				$levels_found_ary = @()  # Initialize collector
				if ($filter_levels_arg_ary_cnt -gt 0) {
					# Write-Host "filtering 'level' column..."
					foreach ($item in $filter_levels_arg_ary) {
						# Write-Host "item = $item"
						$terms = $item.Split('|')
						# Write-Host ("terms = " + ($terms -Join ', '))
						foreach ($term in $terms) {
							$term = $term.Replace('"', '')  # ESSENTIAL!
							# Write-Host ("term = $term, type = " + $term.GetType().Name)
							# Write-Host ("term = " + ($term | ConvertTo-Json))
							if ($term.StartsWith('!')) {
								#$x = $term.TrimStart('!') -as [Int]
								$x = [Int]($term.TrimStart('!'))
								# Write-Host ("!x = $x, type = " + $x.GetType().Name)
								if ($level -ne $x) {
									$levels_found_ary += $true
									break # Exit the loop once a match is found
								}
							} else {
								if ($used_levels -notcontains $term) {
									$used_levels += $term
								}
								#$x = [Convert]::ToInt32($term)
								#$x = $term -as [Int]  # This is the only method that worked without removing the quotes
								$x = [Int]$term
								# Write-Host ("x = $x, type = " + $x.GetType().Name)
								if ($level -eq $x) {
									# Write-Host "match!"
									# Write-Host "levels_found_ary = $levels_found_ary"
									$levels_found_ary += $true
									break # Exit the loop once a match is found
								}
							}
						}
					}
				}

				# ... COLUMN #2 'date'
				#     script does not query this column

				# ... COLUMN #3 'object'
				$objects_found_ary = @()  # Initialize collector
				if ($filter_objects_arg_ary_cnt -gt 0) {
					# Write-Host "filtering 'object' column..."
					foreach ($item in $filter_objects_arg_ary) {
						# Write-Host "item = $item"
						$terms = $item.Split('|')
						# Write-Host ("terms = " + $terms -Join ',')
						foreach ($term in $terms) {
							# Write-Host "term = $term"
							$term = $term.Replace('"', '')  # ESSENTIAL!
							if ($term.StartsWith('!')) {
								$x = $term.TrimStart('!')
								if ($object -notlike "*$x*") {
									$objects_found_ary += $true
									break # Exit the loop once a match is found
								}
							} else {
								if ($used_terms -notcontains $term) {
									$used_terms += $term
								}
								if ($object -like "*$term*") {
									$objects_found_ary += $true
									break # Exit the loop once a match is found
								}
							}
						}
					}
				}

				# ... COLUMN #4 'message'
				$msgs_found_ary = @()  # Initialize collector
				if ($filter_msgs_arg_ary_cnt -gt 0) {
					# Write-Host "filtering 'message' column..."
					foreach ($item in $filter_msgs_arg_ary) {
						# Determine if -FilterMsgs arg is a pure 'OR' request
						if ($filter_msgs_arg_ary_cnt -eq 1 -and $item.IndexOf('|') -ge 0) {
							$pure_or_query_flg = $true
						}
						# Write-Host "item = $item"						
						$terms = $item.Split('|')
						foreach ($term in  $terms) {
							$term = $term.Replace('"', '')  # ESSENTIAL!
							$term = $term.Replace('%2C', ",")  # Char %2C must be used if -FilterMsgs arg contains a term with a comma
							# Write-Host "term = $term"
							if ($term.StartsWith('!')) {
								$x = $term.TrimStart('!')
								if ($message -notlike "*$x*") {
									$msgs_found_ary += $true
									break # Exit the loop once a match is found
								}
							} else {
								if ($used_terms -notcontains $term) {
									$used_terms += $term
								}
								if ($message -like "*$term*") {
									$msgs_found_ary += $true
									
									# For debugging
									# Write-Host ">> MATCH << for term '$term'"
									# Write-Host "msgs_found_ary = $msgs_found_ary"
									
									if ($pure_or_query_flg) {
										if (!$pure_or_query_ht[$term]) {
											$pure_or_query_ht[$term] = @()  # Initialize collector for this term
										}
										$pure_or_query_ht[$term] += $row
									}
									break # Exit the loop once a match is found
								}
							}
						}
					}
				}

				$levels_found_ary_cnt  = $levels_found_ary.count
				$objects_found_ary_cnt = $objects_found_ary.count
				$msgs_found_ary_cnt    = $msgs_found_ary.count

				# For debugging
				# Write-Host "row #$row_num = $row"
				# Write-Host "levels_found_ary_cnt = $levels_found_ary_cnt"
				# Write-Host ("levels_found_ary = " + ($levels_found_ary | ConvertTo-Json))
				# Write-Host "objects_found_ary_cnt = $objects_found_ary_cnt"
				# Write-Host ("objects_found_ary = " + ($objects_found_ary | ConvertTo-Json))
				# Write-Host "msgs_found_ary_cnt = $msgs_found_ary_cnt"
				# Write-Host ("msgs_found_ary = " + ($msgs_found_ary | ConvertTo-Json))

				# APPLY MATCH LOGIC
				if ($filter_levels_arg_ary_cnt -eq 0) {
					$test1 = $true
				} else {
					$test1 = ($levels_found_ary_cnt -eq $filter_levels_arg_ary_cnt)
				}
				if ($filter_objects_arg_ary_cnt -eq 0) {
					$test2 = $true
				} else {
					$test2 = ($objects_found_ary_cnt -eq $filter_objects_arg_ary_cnt)
				}
				if ($filter_msgs_arg_ary_cnt -eq 0) {
					$test3 = $true
				} else {
					$test3 = ($msgs_found_ary_cnt -eq $filter_msgs_arg_ary_cnt)
				}
				
				# For debugging
				# Write-Host "test1|test2|test3 = $test1|$test2|$test3"

				# Add matched row to accumulator
				if ($test1 -and $test2 -and $test3) {
					$matched_rows += $row
				}

			} else {

				$malformed_row_num++
				
				if ($FromConsole -and $batchfile_arg_active_line_num -le 1) {
					if ($malformed_row_num -eq 1) {
						Write-Host ("NOTE: expecting each entry to have 4 tab-delimited columns") -ForegroundColor Red
						Write-Host ("Entries: ") -NoNewLine -ForegroundColor Red
						Write-Host ($newest_rows | ConvertTo-Json)
					}
					Write-Host ("ignoring malformed Entry #" + $row_num + ": " + $row) -ForegroundColor Red
				}
			}
			
			$row_num++
			
		}
	}

	# For debugging
	# Write-Host ("pure_or_query_ht keys cnt = " + $pure_or_query_ht.Keys.count)
	# Write-Host ("pure_or_query_ht keys = " + $pure_or_query_ht.Keys)
	# Write-Host ("pure_or_query_ht = " + ($pure_or_query_ht | ConvertTo-Json))

	$matched_rows_cnt = $matched_rows.Count
	$found_matches_flg = ($matched_rows_cnt -gt 0)

	# Added matched rows to accumulator file to be returned when done executing a -Batchfile
	if ((Test-Path -LiteralPath $batchfile_arg_status_path) -and $found_matches_flg) {
		$content = $matched_rows -Join "`n"
		Add-Content -LiteralPath $batchfile_arg_matches_path -Value $content
	}

	$filter_levels_str = $FilterLevels -Join ','
	$filter_objects_str = $FilterObjects -Join ','
	$filter_msgs_str = $FilterMsgs -Join ','

	# For debugging
	# Write-Host "filter_msgs_str = $filter_msgs_str"

	# SHOW RESULTS
	if ($FromConsole) {
		$offset = ''
		if (!$Quiet) {
			if ($batchfile_arg_active_line_num -gt 0) {
				Write-Host ("`nRESULTS FOR FILTER ACTION #" + $batchfile_arg_active_line_num + ":")
				Write-Host "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
				$offset = '   '
			} else {
				Write-Host "`nFilter Results:"
				Write-Host "~~~~~~~~~~~~~~~"
			}
		}

		if ($using_filter_args_flg -or $using_batchfile_arg_flg) {
			$summary_ary_1 = @()  # Used for console output
			$summary_ary_2 = @()  # Used for Push messages
			if ($filter_levels_arg_ary_cnt -gt 0) {
				$summary_ary_1 += '"level" [ "' + $filter_levels_str + '" ]'
				$summary_ary_2 += '<b>level</b> [ "' + $filter_levels_str + '" ]'
			}
			if ($filter_objects_arg_ary_cnt -gt 0) {
				$summary_ary_1 += '"object" [ "' + $filter_objects_str + '" ]'
				$summary_ary_2 += '<b>object</b> [ "' + $filter_objects_str + '" ]'
			}
			if ($filter_msgs_arg_ary_cnt -gt 0) {
				$summary_ary_1 += '"message" [ "' + $filter_msgs_str + '" ]'
				$summary_ary_2 += '<b>message</b> [ "' + $filter_msgs_str + '" ]'
			}
			
			# For debugging
			# Write-Host ("summary_ary_1 = " + ($summary_ary_1 | ConvertTo-Json))
			# Write-Host ("summary_ary_2 = " + ($summary_ary_2 | ConvertTo-Json))

			# Create a string like,
			# ... 'object' [ 'admin' OR 'DB' ] AND 'message' [ 'login' OR 'logout' AND 'err' ]s
			$summary_str_1 = $summary_ary_1 -Join ' AND '
			$summary_str_1 = $summary_str_1.Replace(',', '" AND "').Replace('|', '" OR "').Replace("!", "NOT ")
			$summary_str_1 = $summary_str_1.Replace('""', '"')
			if ($PSBoundParameters.ContainsKey('TestFile')) {
				Write-Host ("Found $matched_rows_cnt/$newest_rows_cnt rows in -TestFile w/ " + $summary_str_1)
			} else {
				Write-Host ("Found $matched_rows_cnt/$newest_rows_cnt new logfile entries w/ " + $summary_str_1)
			}				
			if ($DevMode) {
				Write-Host ("(dev mode) matched_rows cnt = " + $matched_rows.Count) -ForegroundColor Cyan
				if ($matched_rows.Count -gt 0) {
					Write-Host ("(dev mode) matched_rows = `n" + ($matched_rows | ConvertTo-Json)) -ForegroundColor Cyan
				}
			}
			# CREATE A HTML FORMATTED VERSION FOR PUSH NOTIFICATIONS
			$summary_str_2 = $summary_ary_2 -Join ' AND '
			$summary_str_2 = $summary_str_2.Replace(',', '" AND "').Replace('|', '" OR "').Replace("!", "NOT ")
			$summary_str_2 = $summary_str_2.Replace('""', '"')
							
			# Clean up special characters (required in -BatchFile)
			$summary_str_2 = $summary_str_2.Replace('%2C', ',')
			
		} else {
			#Write-Warning "Nothing to do - please use one or more -Filter* arguments, or the -BatchFile argument"
			Write-Host ("Nothing to do - please use one or more -Filter* arguments, or the -BatchFile argument") -ForegroundColor Yellow
			Write-Host ""
		}

		if ($found_matches_flg) {
			if ($ShowMatches) {
				Write-Host ("--> Entries matched = `n" + ($matched_rows | ConvertTo-Json))
			}
		}			
	}

	# TEST IF WE'RE DONE PROCESSING THE -BatchFile ARGUMENT
	# ... if so, delete the status file
	if (Test-Path -LiteralPath $batchfile_arg_status_path) {
		if ($batchfile_arg_active_line_num -eq $batchfile_arg_lines_cnt) { 
			Remove-Item -LiteralPath $batchfile_arg_status_path

			# For debugging
			# Write-Host "deleted $batchfile_arg_status_path"
		}
	}

	# SEND NOTIFICATION
	if ($found_matches_flg) {		

		if (!$NoPush) {			

			# Build title and end strings
			$end_str = ""  # default
			if (!$Title) {
				$notification_title = "via $ScriptName"  # Default title
			} else {
				$notification_title = $Title
				# Avoid redundant sourcing info in message
				if ($notification_title -notmatch $ScriptName) {
					$end_str = "<br>~~~~~<br>(via $ScriptName)"
				}
			}

			if ($push_service -match "pushover" -or $push_service -match "all") {
				
				$notification_msg = "HEADS-UP:`nFound $matched_rows_cnt/$newest_rows_cnt new BI logfile entries w/ " + $summary_str_2
				
				if ($push_row_matches_disp_max -gt 0) {
					# Highlight the found terms
					$formatted_results_str = `
						FormatLogEntriesForPush `
						-Rows $matched_rows `
						-RowsMax $push_row_matches_disp_max `
						-UsedLevels $used_levels `
						-UsedTerms $used_terms `
						-PureOrQueryHt $pure_or_query_ht
					
					# For debugging
					# Write-Host "formatted_results_str = $formatted_results_str"

					if ($pure_or_query_flg) {
						$str = "most recent matches (1 per found tag):"
					} else {
						$str = "most recent matches ($push_row_matches_disp_max max):"
					}
					
					$notification_msg += "`n~~~~~~~~~~~"
					$notification_msg += "`n<b>$str</b>`n" + $formatted_results_str
					if ($matched_rows_cnt -gt $push_row_matches_disp_max) {
						#$notification_msg += "<br>~~~~~<br>(for all results, see '" + $batchfile_arg_matches_path + "')"
						$notification_msg += "<br>~~~~~<br>(for all results, see '\$scr_working_folder\" + (Split-Path $batchfile_arg_matches_path -leaf) + "')"
					}
					$notification_msg += $end_str
				}
				
				# For debugging
				#Write-Host "po_msg = '$notification_msg'"

				# Returns string - either API response, or "ERR/*"|"WARNING:*" message
				# ... API response is like '@{status=1; request=28f81c23-dc07-4a36-8523-f6428159284c}'
				$response_str = `
					SendPushover `
					-MsgTitle $notification_title `
					-MsgStr $notification_msg `
					-PriorityLevel $notification_priority `
					-Sound $notification_sound `
					-Devices $notification_devices
				
				if ($FromConsole -and !$Quiet) {
					Write-Host ("`n" + $offset + "Pushover Summary")
					Write-Host ($offset + "------------------")
				}

				# For testing - Simulate a bad pushover response
				#$response_str = "ERR/test"
				#$response_str = "WARNING:test"
				
				# for debugging
				#Write-Host "SendPushover() response_str = $response_str"
				
				# Trap error conditions
				if ($response_str.StartsWith("ERR/") -or $response_str.StartsWith("WARNING")) {

					# POST MESSAGE TO BI LOGFILE
					if ($post_to_BI_logfile_flg -and $bi_webserver) {
						$response = PostCustomBiLogEntry `
							-MsgStr "$ScriptName - unexpected response trying to send Pushover notification" `
							-Level 2 `
							-SilentFlg $true
					}

					if ($FromConsole) {
						Write-Host $offset -NoNewLine
						Write-Warning ("Notification not sent")
						Write-Host $offset -NoNewLine
						Write-Host ("ERR/unexpected response from SendPushover(): '" + $response_str + "'") -ForegroundColor Red
					}		
					
				} else {
						
					if ($FromConsole -and !$Quiet) {
						Write-Host ($offset + "Pushover notification sent`n" + $offset  + "API response = "  + $response_str)
					}
					
					<# if ($bg_msgs_on) {
						$bg_msg = "$response_str ($script:ScriptName)"
						PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $script:bg_msgs_on
					} #>
				}
			}
			
			if ($push_service -match "telegram" -or $push_service -match "all") {
				
				$notification_msg = "<b>$notification_title</b><br>HEADS-UP:`nFound $matched_rows_cnt/$newest_rows_cnt new BI logfile entries w/ " + $summary_str_2
				
				if ($push_row_matches_disp_max -gt 0) {

					# Highlight the found terms
					$formatted_results_str = `
						FormatLogEntriesForPush `
						-Rows $matched_rows `
						-RowsMax $push_row_matches_disp_max `
						-UsedLevels $used_levels `
						-UsedTerms $used_terms `
						-PureOrQueryHt $pure_or_query_ht
					
					if ($pure_or_query_flg) {
						$str = "most recent matches (1 per found tag):"
					} else {
						$str = "most recent matches ($push_row_matches_disp_max max):"
					}

					$notification_msg += "`n~~~~~~~~~~~"
					$notification_msg += "`n<b>$str</b>`n" + $formatted_results_str
					if ($matched_rows_cnt -gt $push_row_matches_disp_max) {
						#$notification_msg += "<br>~~~~~<br>(for all results, see '" + $batchfile_arg_matches_path + "')"
						$notification_msg += "<br>~~~~~<br>(for all results, see '\$scr_working_folder\" + (Split-Path $batchfile_arg_matches_path -leaf) + "')"
					}
					$notification_msg += $end_str

				}

				# Returns string - either API response, or "ERR/*"|"WARNING:*" message
				$response_str = `
					SendTelegram `
					-MsgStr $notification_msg

				if ($FromConsole -and !$Quiet) {
					Write-Host ("`n" + $offset + "Telegram Summary:")
					Write-Host ($offset + "------------------")
				}

				# For testing - Simulate a bad pushover response
				#$response_str = "ERR/test"
				#$response_str = "WARNING:test"
				
				# for debugging
				# Write-Host "SendTelegram() response_str = $response_str"
				
				# Trap error conditions
				if ($response_str.StartsWith("ERR/") -or $response_str.StartsWith("WARNING")) {

					# POST MESSAGE TO BI LOGFILE
					if ($post_to_BI_logfile_flg -and $bi_webserver) {
						$response = PostCustomBiLogEntry `
							-MsgStr "$ScriptName - unexpected response trying to send Telegram notification" `
							-Level 2 `
							-SilentFlg $true
					}

					if ($FromConsole) {
						Write-Host $offset -NoNewLine
						Write-Warning ("Notification not sent")
						Write-Host $offset -NoNewLine
						Write-Host ("ERR/unexpected response from SendTelegram(): '" + $response_str + "'") -ForegroundColor Red
					}		
					
				} else {
						
					if ($FromConsole -and !$Quiet) {
						Write-Host ($offset + "Telegram notification sent`n" + $offset  + "API response = "  + $response_str)
					}
					
					<# if ($bg_msgs_on) {
						$bg_msg = "$response_str ($script:ScriptName)"
						PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $script:bg_msgs_on
					} #>
				}
			}

		}

	} else {

		# For debugging
		# msg.exe $server_username /TIME:10 "'$ScriptName' - No issues found"
		
	}

} else {
			
	if ($using_filter_args_flg) {
		Set-Content -LiteralPath $batchfile_arg_no_new_entries_path -Value "no new logfile entries"
	}

	if ($FromConsole) {
		if (!$using_batchfile_arg_flg -and !$using_my_default_filters_flg) {
			#Write-Warning "No new entries detected in the logfile ($bi_logfile_name) since the last check - please retry later"
			Write-Host ("No new entries detected in the logfile ($bi_logfile_name) since the last check - please retry later") -ForegroundColor Yellow
			Write-Host ""
		}
	}
	
	if (!$FromConsole) {
		$rtn_ary = @()
		return $rtn_ary
	}

	# For debugging
	#Write-Host "trying to break out of recursive mode now"
}

# Stop timer
$timer.stop()
$elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
$runtime_elapsed_msec = $elapsed_msec.ToString("0")
# Write-Host ("timer = $runtime_elapsed_msec msec")

# Make strings used in both BI Logfile and CSV log file entries
if (1 -eq 1) {  # Use to collapse this code section while editing

	if ($params_str.Length -eq 0) {
		$script_args_str = "no args"
	} else {
		$script_args_str = $params_str
	}
	# ... Remove unwanted args from the string
	$keepers = @('-FilterLevels', '-FilterObjects', '-FilterMsgs')
	$args_used_ary = @()   # Initialize collector for CSV file
	$args_used_ary2 = @()  # Initialize collector for BI logfile
	foreach ($param in $params_ary) {
		foreach ($keeper in $keepers) {
			if ($param -match $keeper) {
				if ($args_used_ary -notcontains $param) {
					$args_used_ary += $param
					# Shorten arg names to help ensure message length is under 127 cgaracters
					$args_used_ary2 += $param.Replace("-FilterLevels", "-LVL").Replace("-FilterObjects", "-OBJ").Replace("-FilterMsgs", "-MSG")
				}
			}
		}
	}
	$args_used_str = $args_used_ary -Join " "
	$args_used_str2 = $args_used_ary2 -Join " "
	if ($args_used_str.length -eq 0) {
		$args_used_str = "no args"
		$args_used_str2 = "no args"
	}

	# For debugging
	# Write-Host "args_used_str as-is = '$script_args_str'"
	# Write-Host "args_used_str = '$args_used_str'"
	# Write-Host "args_used_str2 = '$args_used_str2'"

	# Get friendly host name
	if ($HostLabels[$ParentProcess]) {
		$script_host_str = $HostLabels[$ParentProcess]
	} else {
		$script_host_str = $ParentProcess
	}

	# For debugging
	# Write-Host "script_host_str = '$script_host_str'"
}

# POST MESSAGE TO BLUE IRIS LOGFILE
if ($found_matches_flg -and !$NoBiLog -and ($using_filter_args_flg -or $using_batchfile_arg_flg)) {

	if ($FromConsole -and !$Quiet) {
		Write-Host ("`n" + $offset + "BI Logfile Post Summary:")
		Write-Host ($offset + "------------------------")
	}
	if ($post_to_BI_logfile_flg -and !$NoBiLog) {

		# For testing
		#$FromConsole = $false
		#$post_to_BI_logfile_in_background_only_flg = $false
		#$post_to_BI_logfile_in_background_only_flg = $true

		# Optionally, messages are posted only if this script is running in the background  
		# (e.g., via Task Scheduler, or Blue Iris actions)
		$doit_flg = $true  # default
		if ($post_to_BI_logfile_in_background_only_flg -and $FromConsole) {
			$doit_flg = $false
		}
		# or, only when some BIlogfile rows are matched
		if ($post_to_BI_logfile_w_matches_only_flg -and !$found_matches_flg) {
			$doit_flg = $false
		}

		# For debugging
		# Write-Host "doit_flg = $doit_flg"	

		if ($doit_flg) {

			if ($bi_webserver) {

				# BUILD BI LOGFILE MESSAGE
				# ... NOTE: function PostCustomBiLogEntry() will trim the message if necessary to meet 127 char max length
				$summary_str = "$matched_rows_cnt/$newest_rows_cnt new rows matched"
				$msg_str = "$ScriptBaseName $summary_str  [$runtime_elapsed_msec ms] [$args_used_str2]"
				
				$response = `
					PostCustomBiLogEntry `
					-MsgStr $msg_str `
					-Level $custom_BI_logfile_level `
					-SilentFlg $true
					
				if ($response -notmatch 'ERR/') {
					# Normal response
					if ($FromConsole -and !$Quiet) {
						if ($DevMode) {
							Write-Host ("(dev mode) Message posted to Blue Iris logfile:`n" + $response) -ForegroundColor Cyan
						} else {
							Write-Host ($offset + "Message posted to Blue Iris logfile")
						}
					}
				} else {
					# Unexpected response
					if ($FromConsole) {
						Write-Host $offset -NoNewLine
						Write-Warning ("Failed to post message to Blue Iris logfile:`n" + $response)
					} else {
						if ($bg_msgs_on) {
							msg.exe $server_username /TIME:5 "From script... $ScriptName`n-----`nFailed to post message to Blue Iris logfile"
						}
					}
				}
			} else {
				if ($FromConsole) {
					Write-Host $offset -NoNewLine
					Write-Warning ("No message posted to the Blue Iris logfile; unable to retrieve web server address from the registry")
				}
			}
		} else {
			if ($FromConsole) {
				if ($post_to_BI_logfile_in_background_only_flg) {
					Write-Host $offset -NoNewLine
					Write-Warning ("No message posted to the Blue Iris logfile (not running in background - per user-settings-file)")
				} elseif ($post_to_BI_logfile_w_matches_only_flg -and !$found_matches_flg) {
					Write-Host $offset -NoNewLine
					Write-Warning ("No message posted to the Blue Iris logfile (no matches found to report - per user-settings-file)")
				} else {
					Write-Host $offset -NoNewLine
					Write-Warning ("No message posted to the Blue Iris logfile")
				}
			}
		}
	} else {
		if ($FromConsole -and !$Quiet) {
			Write-Host ($offset + "No message posted to the Blue Iris logfile")
		}
	}
}

# HANDLE SCRIPT LOGGING (CSV FILE)
if ($log_to_csv_logfile_flg) {

	$doit_flg = $true
	if ($log_only_events_to_csv_logfile_flg -and !$memleak_event_flg) {	
		$doit_flg = $false
	}
	if ($doit_flg) {	
		# Prep values for logging
		$csvfilespec = $app_path + "\" + $ScriptBaseName + "_log_" + $ScriptVersion + ".csv"
		$timestamp = Get-Date -Format "yyyy/MM/dd_HH:mm:ss"
		if ($found_matches_flg -and !$NoPush) {
			$pushsent_str = $push_service
		} else {
			$pushsent_str = "no push"
		}

		$csv_response = `
			LogScriptResultsToCsv `
			-CsvFilespec $csvfilespec `
			-Timestamp $timestamp `
			-NewestRowsCnt $newest_rows_cnt `
			-MatchedRowsCnt $matched_rows_cnt `
			-PushSent $pushsent_str `
			-ScriptMsec $runtime_elapsed_msec `
			-ScriptHost $script_host_str `
			-ArgsUsedStr $args_used_str

		# For debugging
		# Write-Host "LogScriptResultsToCsv csv_response = $csv_response"
		
		if ($csv_response) {
			if ($csv_response.StartsWith("ERR/")) {
				if ($FromConsole) {
					Write-Warning "unable to update script's CSV log (Reason: $csv_response)"
				}
			}
		} else {
			if ($FromConsole) {
				Write-Warning "unable to update script's CSV log (Reason unknown)"
			}
		}
	}
}

if ($FromConsole) {
	Write-Host ""
} else {
	return $matched_rows
}


